import asyncio
import uuid
from datetime import datetime
from typing import Any
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import config
import los_files
import mox_helpers.payloads as mox_payloads
import pydantic
import uuids
from mox_helpers.mox_helper import create_mox_helper
from mox_helpers.mox_helper import ElementNotFound
from pydantic import Field

from os2mo_data_import.mox_data_types import Facet
from os2mo_data_import.mox_data_types import Itsystem


class StamCSV(pydantic.BaseModel):
    @staticmethod
    def get_filename() -> str:
        raise NotImplementedError("must be implemented by subclass")

    @staticmethod
    def get_facet_bvn() -> str:
        raise NotImplementedError("must be implemented by subclass")

    @property
    def bvn(self) -> str:
        raise NotImplementedError("must be implemented by subclass")

    @property
    def title(self) -> str:
        raise NotImplementedError("must be implemented by subclass")

    @property
    def class_uuid(self) -> uuid.UUID:
        raise NotImplementedError("must be implemented by subclass")


class Engagementstype(StamCSV):
    engagement_type_uuid: uuid.UUID = Field(alias="EngagementstypeUUID")
    engagement_type: str = Field(alias="Engagementstype")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Engagementstype.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "engagement_type"

    @property
    def bvn(self) -> str:
        return self.engagement_type

    @property
    def title(self) -> str:
        return self.engagement_type

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.engagement_type_uuid


class Enhedstype(StamCSV):
    org_unit_type_uuid: uuid.UUID = Field(alias="EnhedstypeUUID")
    org_unit_type: str = Field(alias="Enhedstype")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Enhedstype.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "org_unit_type"

    @property
    def bvn(self) -> str:
        return self.org_unit_type

    @property
    def title(self) -> str:
        return self.org_unit_type

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.org_unit_type_uuid


class ITSystem(StamCSV):
    it_system_uuid: uuid.UUID = Field(alias="ITSystemUUID")
    name: str = Field(alias="Name")
    user_key: str = Field(alias="Userkey")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_ITSystem.csv"


class Stillingsbetegnelse(StamCSV):
    job_function_uuid: uuid.UUID = Field(alias="StillingBetUUID")
    job_function: str = Field(alias="Stillingsbetegnelse")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Stillingsbetegnelse.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "engagement_job_function"

    @property
    def bvn(self) -> str:
        return self.job_function

    @property
    def title(self) -> str:
        return self.job_function

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.job_function_uuid


class BVNStillingsbetegnelse(StamCSV):
    job_function_uuid: uuid.UUID = Field(alias="BVNStillingBetUUID")
    job_function: str = Field(alias="BVNStillingsbetegnelse")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_BVN_Stillingsbetegnelse.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "engagement_job_function_bvn"

    @property
    def bvn(self) -> str:
        return self.job_function

    @property
    def title(self) -> str:
        return self.job_function

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.job_function_uuid


class Lederansvar(StamCSV):
    responsibility_uuid: uuid.UUID = Field(alias="LederansvarUUID")
    responsibility: str = Field(alias="Lederansvar")
    loadtime: datetime = Field(alias="Loadtime")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Lederansvar.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "responsibility"

    @property
    def bvn(self) -> str:
        return self.responsibility

    @property
    def title(self) -> str:
        return self.responsibility

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.responsibility_uuid


class Lederniveau(StamCSV):
    level_uuid: uuid.UUID = Field(alias="LederniveauUUID")
    level: str = Field(alias="Lederniveau")
    loadtime: datetime = Field(alias="Loadtime")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Lederniveau.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "manager_level"

    @property
    def bvn(self) -> str:
        return self.level

    @property
    def title(self) -> str:
        return self.level

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.level_uuid


class Ledertype(StamCSV):
    type_uuid: uuid.UUID = Field(alias="LedertypeUUID")
    type: str = Field(alias="Ledertype")
    loadtime: datetime = Field(alias="Loadtime")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Ledertype.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "manager_type"

    @property
    def bvn(self) -> str:
        return self.type

    @property
    def title(self) -> str:
        return self.type

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.type_uuid


class Tilknytningsrolle(StamCSV):
    association_type_uuid: uuid.UUID = Field(alias="TilknytningsrolleUUID")
    role: str = Field(alias="Tilknytningsrolle")
    loadtime: datetime = Field(alias="Loadtime")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Tilknytningsrolle.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "association_type"

    @property
    def bvn(self) -> str:
        return self.role

    @property
    def title(self) -> str:
        return self.role

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.association_type_uuid


StamCSVType = Union[
    Type[Engagementstype],
    Type[Enhedstype],
    Type[ITSystem],
    Type[Stillingsbetegnelse],
    Type[BVNStillingsbetegnelse],
    Type[Lederansvar],
    Type[Lederniveau],
    Type[Ledertype],
    Type[Tilknytningsrolle],
]


class StamImporter:
    """Import all classes and IT systems from CSV files.
    Imports based on predetermined UUIDs, so these functions should be idempotent.
    Only imports if file has been modified after `last_import`.
    """

    def __init__(self, last_import: datetime):
        super().__init__()
        self.last_import = last_import

    async def run(self):
        print("Starting STAM import")
        tasks = [
            self.handle_engagementstype(),
            self.handle_enhedstype(),
            self.handle_itsystem(),
            self.handle_stillingsbetegnelse(),
            self.handle_bvn_stillingsbetegnelse(),
            self.handle_lederansvar(),
            self.handle_lederniveau(),
            self.handle_ledertype(),
            self.handle_tilknytningsrolle(),
        ]
        await asyncio.gather(*tasks)
        print("STAM import done")

    async def handle_engagementstype(self):
        """
        Process the external 'engagement type' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Engagementstype)
        return await self._create_classes_from_csv(Engagementstype, rows)

    async def handle_enhedstype(self):
        """
        Process the external 'unit type' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Enhedstype)
        return await self._create_classes_from_csv(Enhedstype, rows)

    async def handle_itsystem(self):
        """
        Process the external 'it system' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(ITSystem)
        if rows is None:
            return

        settings = config.get_config()
        mox_helper = await create_mox_helper(settings.mox_base)
        for row in rows:
            it_system = Itsystem(
                system_name=row.name,  # type: ignore
                user_key=row.user_key,  # type: ignore
            )
            it_system.organisation_uuid = str(uuids.ORG_UUID)
            json = it_system.build()
            await mox_helper.insert_organisation_itsystem(
                json,
                str(row.it_system_uuid),  # type: ignore
            )

    async def handle_stillingsbetegnelse(self):
        """
        Process the external 'job_function' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Stillingsbetegnelse)
        return await self._create_classes_from_csv(Stillingsbetegnelse, rows)

    async def handle_bvn_stillingsbetegnelse(self):
        """
        Process the external 'BVNStillingsbetegnelse' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(BVNStillingsbetegnelse)
        return await self._create_classes_from_csv(BVNStillingsbetegnelse, rows)

    async def handle_lederansvar(self):
        """
        Process the external 'Lederansvar' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Lederansvar)
        return await self._create_classes_from_csv(Lederansvar, rows)

    async def handle_lederniveau(self):
        """
        Process the external 'Lederniveau' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Lederniveau)
        return await self._create_classes_from_csv(Lederniveau, rows)

    async def handle_ledertype(self):
        """
        Process the external 'Ledertype' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Ledertype)
        return await self._create_classes_from_csv(Ledertype, rows)

    async def handle_tilknytningsrolle(self):
        """
        Process the external 'Tilknytningsrolle' file and create objects if file is
        newer than last import.
        """
        rows = self._load_csv_if_newer(Tilknytningsrolle)
        return await self._create_classes_from_csv(Tilknytningsrolle, rows)

    def _load_csv_if_newer(
        self, csv_class: StamCSVType
    ) -> Union[List[StamCSVType], None]:
        filename = csv_class.get_filename()
        fileset = los_files.get_fileset_implementation()
        try:
            modified_datetime = fileset.get_modified_datetime(filename)
        except ValueError:
            # Raised by `fileset.get_modified_datetime` if file could not be
            # found.
            return None
        else:
            if modified_datetime <= self.last_import:
                return None
            return los_files.read_csv(filename, csv_class)

    async def _create_classes_from_csv(
        self, csv_class: StamCSVType, rows: Optional[List[StamCSVType]] = None
    ) -> None:
        if rows is None:
            return

        settings = config.get_config()
        mox_helper = await create_mox_helper(settings.mox_base)

        facet_uuid = await self._get_or_create_facet(csv_class, mox_helper)

        # Find class UUIDs currently in facet
        facet_class_uuids = list(
            map(
                uuid.UUID,
                set(
                    await mox_helper._search(
                        "klassifikation",
                        "klasse",
                        {"vilkaarligrel": str(facet_uuid)},
                    )
                ),
            )
        )

        # Find rows to insert (classes in CSV but not yet in LoRa)
        rows_class_uuids = set(row.class_uuid for row in rows)
        rows_to_insert = [
            row for row in rows if row.class_uuid not in facet_class_uuids
        ]

        # Create classes in LoRa
        for row in rows_to_insert:
            klasse = mox_payloads.lora_klasse(
                bvn=row.bvn,
                title=row.title,
                facet_uuid=str(facet_uuid),
                org_uuid=str(uuids.ORG_UUID),
                scope="TEXT",
            )
            await mox_helper.insert_klassifikation_klasse(klasse, str(row.class_uuid))
            print("Created LoRa class %r" % row.class_uuid)

        # Find classes to terminate (classes in LoRa but no longer in CSV file)
        class_uuids_to_terminate = [
            class_uuid
            for class_uuid in facet_class_uuids
            if class_uuid not in rows_class_uuids
        ]

        # Terminate classes in LoRa by setting them as unpublished
        unpublish_payload = self._get_lora_unpublish_payload()
        for class_uuid in class_uuids_to_terminate:
            try:
                await mox_helper._update(
                    "klassifikation",
                    "klasse",
                    str(class_uuid),
                    unpublish_payload,
                )
            except KeyError:  # this class is already unpublished
                print("LoRa class %r was already unpublished" % class_uuid)
            else:
                print("Unpublished LoRa class %r" % class_uuid)

    @staticmethod
    def _get_lora_unpublish_payload():
        return {
            "attributter": {
                "klasseegenskaber": [
                    {
                        "brugervendtnoegle": "-",
                        "titel": "-",
                        "virkning": {
                            "from": "1910-01-01 00:00:00+01",
                            "to": "infinity",
                        },
                    }
                ]
            },
            "tilstande": {
                "klassepubliceret": [
                    {
                        "publiceret": "IkkePubliceret",
                        "virkning": {
                            "from": "1910-01-01 00:00:00+01",
                            "to": "infinity",
                        },
                    }
                ]
            },
        }

    @staticmethod
    async def _get_or_create_facet(csv_class: StamCSVType, mox_helper: Any) -> str:
        facet_bvn = csv_class.get_facet_bvn()

        try:
            # Get facet by its BVN
            facet_uuid = await mox_helper.read_element_klassifikation_facet(
                bvn=facet_bvn
            )
        except ElementNotFound:
            # Facet does not yet exist, create it based on the BVN
            print(f"LoRa facet {facet_bvn!r} not found, creating ...")
            facet_uuid = uuids.uuid_gen(facet_bvn)
            org_uuid = (await mox_helper.read_all_organisation_organisation())[0]
            kls_uuid = (await mox_helper.read_all_klassifikation_klassifikation())[0]
            facet = Facet(
                user_key=facet_bvn,
                uuid=facet_uuid,
                organisation_uuid=org_uuid,
                klassifikation_uuid=kls_uuid,
            )
            await mox_helper.insert_klassifikation_facet(facet.build(), facet_uuid)

        return facet_uuid



from csv import DictReader
from dataclasses import dataclass
from itertools import groupby
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple
from uuid import UUID

from more_itertools import flatten
from pydantic import BaseModel, PrivateAttr

from os2mo_data_import.Clients.MO.model import OrgUnit
from os2mo_data_import.util import generate_uuid


@dataclass(frozen=True)
class PartialManager:
    engagement_user_key: str
    org_unit_uuid: UUID


class RawOrgUnit(BaseModel):
    Business_Level_2: str
    Business_Level_3: str
    Business_Level_4: str
    Business_Level_5: str
    Business_Level_6: str
    Business_Level_7: str
    Business_Level_8: str
    Business_Level_9: str
    Org_Unit_No: int
    Manager_Init: str
    Rowcounter: int
    __path: List[str] = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self.__path = list(
            filter(
                bool,
                [
                    self.Business_Level_2,
                    self.Business_Level_3,
                    self.Business_Level_4,
                    self.Business_Level_5,
                    self.Business_Level_6,
                    self.Business_Level_7,
                    self.Business_Level_8,
                    self.Business_Level_9,
                ],
            )
        )

    def level(self) -> int:
        """
        used to sort by level
        :return:
        """

        return len(self.__path)

    @staticmethod
    def __spaces_to_underscores(obj: Dict[str, Any]) -> Dict[str, Any]:
        return dict(
            map(lambda keyval: (keyval[0].replace(" ", "_"), keyval[1]), obj.items())
        )

    @classmethod
    def from_dict_with_spaces_in_keys(cls, obj: Dict[str, Any]) -> "RawOrgUnit":
        """
        should be replaced with Field(alias="...") but I didn't know about that

        :param obj:
        :return:
        """
        obj = cls.parse_obj(cls.__spaces_to_underscores(obj))
        return obj

    def __get_uuid(self) -> UUID:
        return generate_uuid(str(self.__path))

    def __get_parent_uuid(self) -> Optional[UUID]:
        if not self.__path[1:]:
            return None
        return generate_uuid(str(self.__path[:-1]))

    def to_mo_org_unit(
            self, org_unit_type_uuid: UUID, org_unit_level_uuid: UUID
    ) -> OrgUnit:
        return OrgUnit.from_simplified_fields(
            uuid=self.__get_uuid(),
            user_key=str(self.Org_Unit_No),
            name=self.__path[-1],
            parent_uuid=self.__get_parent_uuid(),
            org_unit_type_uuid=org_unit_type_uuid,
            org_unit_level_uuid=org_unit_level_uuid,
        )

    def get_partial_manager(self) -> PartialManager:
        return PartialManager(
            engagement_user_key=self.Manager_Init, org_unit_uuid=self.__get_uuid()
        )


def read_csv(
        path: Path, org_unit_type_uuid: UUID, org_unit_level_uuid: UUID
) -> Tuple[List[List[OrgUnit]], List[PartialManager]]:
    with path.open("r") as file:
        raw_org_units = list(
            map(RawOrgUnit.from_dict_with_spaces_in_keys, DictReader(file))
        )
        raw_org_units = list(sorted(raw_org_units, key=lambda x: x.level()))

        raw_org_units = [
            list(group)
            for _, group in groupby(raw_org_units, key=lambda x: x.level())
        ]

        ous = [
            list(
                map(
                    lambda unit: unit.to_mo_org_unit(
                        org_unit_type_uuid=org_unit_type_uuid,
                        org_unit_level_uuid=org_unit_level_uuid,
                    ),
                    group,
                )
            )
            for group in raw_org_units
        ]
        # shitty transformation to list of single-element-lists, shouldn't need this,
        # but I can't get it to work without
        ret = [[ou] for ou in flatten(ous)]

        return ret, list(
            map(lambda x: x.get_partial_manager(), flatten(raw_org_units))
        )


if __name__ == "__main__":
    with Path("/home/mw/gir/gir_orsted_hierarki_linje.csv").open("r") as file:
        RawOrgUnits = map(RawOrgUnit.from_dict_with_spaces_in_keys, DictReader(file))
        for x in RawOrgUnits:
            print(x)
            exit()



from unittest import TestCase

from parameterized import parameterized

from ..ad_writer import MORESTSource
from .mocks import MO_UUID
from .mocks import MockMoraHelper


class _MockMoraHelperMultipleEmails(MockMoraHelper):
    def get_e_addresses(self, e_uuid, scope=None):
        # Duplicate the single-element list returned by `get_e_adddresses`
        return super().get_e_addresses(e_uuid, scope) * 2


class _MockMoraHelperEmptyEmails(MockMoraHelper):
    def get_e_addresses(self, e_uuid, scope=None):
        return []


class _TestableMORESTSource(MORESTSource):
    def __init__(self, helper):
        self.helper = helper


class TestMORESTSource(TestCase):
    @parameterized.expand(
        [
            # Regression test: `MORESTSource.get_email_address` crashed on employees with
            # multiple email addresses.
            # See: #49971
            (_MockMoraHelperMultipleEmails, {"value": "address-value"}),
            # Regression test: `MORESTSource.get_email_address` crashed on employees with
            # no email addresses (empty list of addresses.)
            # See: #50325
            (_MockMoraHelperEmptyEmails, {}),
        ]
    )
    def test_get_email_address(self, mock_cls: MockMoraHelper, expected_addr: dict):
        instance = _TestableMORESTSource(mock_cls("cpr"))
        actual_addr = instance.get_email_address(MO_UUID)
        self.assertDictEqual(actual_addr, expected_addr)



import datetime
from collections import OrderedDict
from operator import itemgetter
from unittest import TestCase
from unittest.mock import call
from unittest.mock import MagicMock

import hypothesis.strategies as st
from hypothesis import given
from more_itertools import unzip

from ..common import MOPrimaryEngagementUpdater
from integrations.ad_integration.utils import AttrDict


def engagements_at_date(date, engagements):
    """Filter engagements to only show ones valid at date.

    This is supposed to emulate how MO handles the `at` parameter.

    Args:
        date: The date to check validities against.
        engagements: The list of engagements to filter

    Returns:
        Filtered list of engagements.
    """

    def check_from_date(engagement):
        from_date = datetime.datetime.strptime(
            engagement["validity"]["from"], "%Y-%m-%d"
        )
        return from_date <= date

    def check_to_date(engagement):
        if engagement["validity"]["to"] is None:
            return True
        to_date = datetime.datetime.strptime(engagement["validity"]["to"], "%Y-%m-%d")
        return date <= to_date

    return list(filter(check_from_date, filter(check_to_date, engagements)))


class MOPrimaryEngagementUpdaterTest(MOPrimaryEngagementUpdater):
    def __init__(self, *args, **kwargs):
        self.morahelper_mock = MagicMock()
        self.morahelper_mock.read_organisation.return_value = "org_uuid"

        super().__init__(*args, **kwargs)

    def _get_mora_helper(self, mora_base):
        return self.morahelper_mock

    def _find_primary_types(self):
        primary_dict = {
            "fixed_primary": "fixed_primary_uuid",
            "primary": "primary_uuid",
            "non_primary": "non_primary_uuid",
            "special_primary": "special_primary_uuid",
        }
        primary_list = [
            primary_dict["fixed_primary"],
            primary_dict["primary"],
            primary_dict["special_primary"],
        ]
        return primary_dict, primary_list

    def _find_primary(self, mo_engagements):
        return mo_engagements[0]["uuid"]


class Test_check_user(TestCase):
    """Test the check_user functions."""

    def setUp(self):
        self.updater = MOPrimaryEngagementUpdaterTest({"mora.base": "mora_base_url"})

    def test_create(self):
        """Test that setUp runs without using it for anything."""
        pass

    @given(
        st.lists(
            st.sampled_from(
                [
                    "primary_uuid",
                    "fixed_primary_uuid",
                    "special_primary_uuid",
                    "non_primary_uuid",
                    "unrelated_uuid",
                ]
            )
        )
    )
    def test_check_user_non_overlapping(self, engagements):
        """Test the result of running _check_user on non-overlapping engagements.

        Args:
            engagements: A list of primary uuids, these are used to create a
                list of actual engagements, with non-overlapping validities.
        """
        # Create a mapping from 'from date' to engagement.
        # 'from_date' is mocked to be (2930 + list index)
        engagement_map = {
            datetime.datetime(2930 + i, 1, 1): engagement
            for i, engagement in enumerate(engagements)
        }
        # This effectively mocks engagement validities, as:
        #   from: (2930 + list index)
        #   to: (2930 + list index + 1)
        # or in the case of the last engagement until 9999-12-30
        cut_dates = list(engagement_map.keys()) + [
            datetime.datetime(9999, 12, 30, 0, 0)
        ]
        self.updater.morahelper_mock.find_cut_dates.return_value = cut_dates

        # As engagements are made non-overlapping, we will always return only one,
        # namely the one found by lookup in our engagement_map
        self.updater._read_engagement = lambda user_uuid, date: [
            {"primary": {"uuid": engagement_map[date]}}
        ]
        check_filters = [
            # Filter out special primaries
            lambda user_uuid, eng: eng["primary"]["uuid"]
            != "special_primary_uuid"
        ]

        def gen_expected(date):
            """Due to non-overlapping 1 or 0 will be returned for either."""
            uuid = engagement_map.get(date)
            count = 1 if uuid in self.updater.primary else 0
            special_count = 1 if uuid == "special_primary_uuid" else 0
            return 1, count, count - special_count

        self.assertEqual(
            self.updater._check_user(check_filters, "user_uuid"),
            {date: gen_expected(date) for date in cut_dates[:-1]},
        )

    def engagements_fixture(self):
        """Engagement fixture for testing overlapping engagements."""
        engagements = [
            {
                "validity": {
                    "from": "1931-1-1",
                    "to": "1950-1-1",
                },
                "uuid": "primary_uuid",
            },
            {
                "validity": {
                    "from": "1939-9-1",
                    "to": "1945-9-2",
                },
                "uuid": "fixed_primary_uuid",
            },
            {
                "validity": {
                    "from": "1949-1-1",
                    "to": None,
                },
                "uuid": "special_primary_uuid",
            },
        ]
        return engagements

    def test_mora_cut_dates(self):
        """Test that mora cut-dates work as expected."""
        from os2mo_helpers.mora_helpers import MoraHelper

        mora_helper = MoraHelper()
        mora_helper.read_user_engagement = MagicMock()
        mora_helper.read_user_engagement.return_value = self.engagements_fixture()

        cut_dates = mora_helper.find_cut_dates("user_uuid")

        # Expected data derived from engagements_fixture
        self.assertEqual(
            cut_dates,
            [
                datetime.datetime(1931, 1, 1),
                datetime.datetime(1939, 9, 1),
                datetime.datetime(1945, 9, 3),  # +1
                datetime.datetime(1949, 1, 1),
                datetime.datetime(1950, 1, 2),  # +1
                datetime.datetime(9999, 12, 30, 0, 0),
            ],
        )

    def test_engagements_at_date(self):
        """Test that engagements_at_date works as expected."""
        engagements = self.engagements_fixture()
        # Expected data derived from engagements_fixture
        engagements_at_date_tests = {
            datetime.datetime(1930, 1, 1): [],
            datetime.datetime(1931, 2, 1): ["primary_uuid"],
            datetime.datetime(1938, 10, 1): ["primary_uuid"],
            datetime.datetime(1939, 10, 1): ["primary_uuid", "fixed_primary_uuid"],
            datetime.datetime(1945, 8, 1): ["primary_uuid", "fixed_primary_uuid"],
            datetime.datetime(1946, 8, 1): ["primary_uuid"],
            datetime.datetime(1948, 2, 1): ["primary_uuid"],
            datetime.datetime(1949, 2, 1): ["primary_uuid", "special_primary_uuid"],
            datetime.datetime(1951, 2, 1): ["special_primary_uuid"],
        }
        for date, expected in engagements_at_date_tests.items():
            filtered_engagements = engagements_at_date(date, engagements)
            engagement_uuids = list(map(itemgetter("uuid"), filtered_engagements))
            self.assertEqual(engagement_uuids, expected)

    def test_check_user_overlapping(self):
        """Test the result of running _check_user on overlapping engagements."""
        # See test_engagement_at_date for details
        engagements = self.engagements_fixture()
        self.updater._read_engagement = lambda user_uuid, date: [
            {"primary": {"uuid": engagement["uuid"]}}
            for engagement in engagements_at_date(date, engagements)
        ]

        # See test_mora_cut_dates for details
        cut_dates = [
            datetime.datetime(1931, 1, 1),
            datetime.datetime(1939, 9, 1),
            datetime.datetime(1945, 9, 3),  # +1
            datetime.datetime(1949, 1, 1),
            datetime.datetime(1950, 1, 2),  # +1
            datetime.datetime(9999, 12, 30, 0, 0),
        ]
        self.updater.morahelper_mock.find_cut_dates.return_value = cut_dates

        check_filters = [
            # Filter out special primaries
            lambda user_uuid, eng: eng["primary"]["uuid"]
            != "special_primary_uuid"
        ]

        self.assertEqual(
            self.updater._check_user(check_filters, "user_uuid"),
            {
                # Only primary_uuid
                datetime.datetime(1931, 1, 1, 0, 0): (1, 1, 1),
                # Both primary_uuid and fixed_primary_uuid
                datetime.datetime(1939, 9, 1, 0, 0): (2, 2, 2),
                # Only primary_uuid
                datetime.datetime(1945, 9, 3, 0, 0): (1, 1, 1),
                # Both primary_uuid and special_primary_uuid
                datetime.datetime(1949, 1, 1, 0, 0): (2, 2, 1),
                # Only special_primary_uuid
                datetime.datetime(1950, 1, 2, 0, 0): (1, 1, 0),
            },
        )

    def test_check_user_outputter(self):
        fixture_data = [
            (datetime.datetime(1931, 1, 1, 0, 0), (2, 0, 0)),
            (datetime.datetime(1932, 1, 1, 0, 0), (2, 1, 1)),
            (datetime.datetime(1933, 1, 1, 0, 0), (2, 2, 0)),
            (datetime.datetime(1934, 1, 1, 0, 0), (2, 2, 1)),
            (datetime.datetime(1935, 1, 1, 0, 0), (2, 2, 2)),
        ]
        # It does not normally return an ordered dict, but for testing we want a
        # consistent order.
        self.updater._check_user = lambda check_filters, user_uuid: OrderedDict(
            fixture_data
        )

        outputter, strings, user_uuids, dates = unzip(
            self.updater._check_user_outputter([], "user_uuid")
        )
        from ..common import noop, logger

        self.assertEqual(
            list(outputter), [print, noop, logger.info, logger.info, print]
        )
        self.assertEqual(
            list(strings),
            [
                "No primary",
                "",
                "All primaries are special",
                "Only one non-special primary",
                "Too many primaries",
            ],
        )
        self.assertEqual(list(user_uuids), ["user_uuid"] * 5)
        self.assertEqual(list(dates), list(map(itemgetter(0), fixture_data)))

        outputter, final_strings = unzip(
            self.updater._check_user_strings([], "user_uuid")
        )
        self.assertEqual(
            list(final_strings),
            [
                "No primary for user_uuid at 1931-01-01",
                " for user_uuid at 1932-01-01",
                "All primaries are special for user_uuid at 1933-01-01",
                "Only one non-special primary for user_uuid at 1934-01-01",
                "Too many primaries for user_uuid at 1935-01-01",
            ],
        )


class Test_recalculate_user(TestCase):
    """Test the recalculate_user functions."""

    def setUp(self):
        self.updater = MOPrimaryEngagementUpdaterTest({"mora.base": "mora_base_url"})
        self.updater.morahelper_mock._mo_post.return_value = AttrDict(
            {
                "status_code": 200,
            }
        )
        self.updater._ensure_primary = MagicMock(wraps=self.updater._ensure_primary)

    def test_create(self):
        """Test that setUp runs without using it for anything."""
        pass

    def test_recalculate_no_engagements(self):
        """Test that no engagements mean no changes and no attempted updates."""
        self.assertEqual(self.updater.recalculate_user("user_uuid"), {"user_uuid": 0})
        self.updater._ensure_primary.assert_not_called()

    @given(st.sampled_from(["primary_uuid", "fixed_primary_uuid"]))
    def test_recalculate_single_engagement_already_primary(self, old_primary):
        """Test that a primary engagement is still primary after recalculate."""
        cut_dates = [
            datetime.datetime(2930, 1, 1),
            datetime.datetime(9999, 12, 30, 0, 0),
        ]
        self.updater.morahelper_mock.find_cut_dates.return_value = cut_dates

        engagement = {"uuid": "engagement_uuid", "primary": {"uuid": old_primary}}
        self.updater._read_engagement = lambda user_uuid, date: [engagement]

        self.assertEqual(self.updater.recalculate_user("user_uuid"), {"user_uuid": 0})
        self.updater._ensure_primary.assert_called_with(
            engagement, old_primary, {"from": "2930-01-01", "to": None}
        )
        self.updater.morahelper_mock._mo_post.assert_not_called()

    @given(st.sampled_from(["non_primary_uuid", "unrelated_uuid"]))
    def test_recalculate_single_engagement_becoming_primary(self, old_primary):
        """Test that a non-primary engagement becomes primary after recalculate."""
        cut_dates = [
            datetime.datetime(2930, 1, 1),
            datetime.datetime(9999, 12, 30, 0, 0),
        ]
        self.updater.morahelper_mock.find_cut_dates.return_value = cut_dates

        engagement = {"uuid": "engagement_uuid", "primary": {"uuid": old_primary}}
        self.updater._read_engagement = lambda user_uuid, date: [engagement]

        self.assertEqual(self.updater.recalculate_user("user_uuid"), {"user_uuid": 1})
        self.updater._ensure_primary.assert_called_with(
            engagement, "primary_uuid", {"from": "2930-01-01", "to": None}
        )
        self.updater.morahelper_mock._mo_post.assert_called_with(
            "details/edit",
            {
                "type": "engagement",
                "uuid": "engagement_uuid",
                "data": {
                    "primary": {"uuid": "primary_uuid"},
                    "validity": {"from": "2930-01-01", "to": None},
                },
            },
        )

    def test_recalculate_multiple_engagements(self):
        """Test that non-primary engagements yield one primary after recalculate.

        Note: which one is subject to the _find_primary method, the test one simply
              picks the first one in the provided list.
        """
        engagements = [
            {"uuid": "engagement_uuid_1", "primary": {"uuid": "non_primary_uuid"}},
            {"uuid": "engagement_uuid_2", "primary": {"uuid": "non_primary_uuid"}},
        ]

        cut_dates = [
            datetime.datetime(2930, 1, 1),
            datetime.datetime(9999, 12, 30, 0, 0),
        ]
        self.updater.morahelper_mock.find_cut_dates.return_value = cut_dates

        self.updater._read_engagement = lambda user_uuid, date: engagements

        self.assertEqual(self.updater.recalculate_user("user_uuid"), {"user_uuid": 1})
        self.updater._ensure_primary.assert_has_calls(
            [
                call(
                    engagements[0], "primary_uuid", {"from": "2930-01-01", "to": None}
                ),
                call(
                    engagements[1],
                    "non_primary_uuid",
                    {"from": "2930-01-01", "to": None},
                ),
            ]
        )
        # Only one call, as non_primary is already non_primary
        self.updater.morahelper_mock._mo_post.assert_called_with(
            "details/edit",
            {
                "type": "engagement",
                "uuid": "engagement_uuid_1",
                "data": {
                    "primary": {"uuid": "primary_uuid"},
                    "validity": {"from": "2930-01-01", "to": None},
                },
            },
        )

    def test_recalculate_multiple_engagements_wrong_primary(self):
        """Test that opposite primary engagements yield two after changes.

        Note: which one is subject to the _find_primary method, the test one simply
              picks the first one in the provided list.
        """
        engagements = [
            {"uuid": "engagement_uuid_1", "primary": {"uuid": "non_primary_uuid"}},
            {"uuid": "engagement_uuid_2", "primary": {"uuid": "primary_uuid"}},
        ]

        cut_dates = [
            datetime.datetime(2930, 1, 1),
            datetime.datetime(9999, 12, 30, 0, 0),
        ]
        self.updater.morahelper_mock.find_cut_dates.return_value = cut_dates

        self.updater._read_engagement = lambda user_uuid, date: engagements

        self.assertEqual(self.updater.recalculate_user("user_uuid"), {"user_uuid": 2})
        self.updater._ensure_primary.assert_has_calls(
            [
                call(
                    engagements[0], "primary_uuid", {"from": "2930-01-01", "to": None}
                ),
                call(
                    engagements[1],
                    "non_primary_uuid",
                    {"from": "2930-01-01", "to": None},
                ),
            ]
        )
        # Two calls, as primary is flipped for each
        self.updater.morahelper_mock._mo_post.assert_has_calls(
            [
                call(
                    "details/edit",
                    {
                        "type": "engagement",
                        "uuid": "engagement_uuid_1",
                        "data": {
                            "primary": {"uuid": "primary_uuid"},
                            "validity": {"from": "2930-01-01", "to": None},
                        },
                    },
                ),
                call(
                    "details/edit",
                    {
                        "type": "engagement",
                        "uuid": "engagement_uuid_2",
                        "data": {
                            "primary": {"uuid": "non_primary_uuid"},
                            "validity": {"from": "2930-01-01", "to": None},
                        },
                    },
                ),
            ]
        )

    def test_recalculate_multiple_engagements_fixed_primary(self):
        """Test that fixed primaries overrule normal primary calculation.

        Note: which one is subject to the _find_primary method, the test one simply
              picks the first one in the provided list.
        """
        engagements = [
            {"uuid": "engagement_uuid_1", "primary": {"uuid": "non_primary_uuid"}},
            {"uuid": "engagement_uuid_2", "primary": {"uuid": "fixed_primary_uuid"}},
        ]

        cut_dates = [
            datetime.datetime(2930, 1, 1),
            datetime.datetime(9999, 12, 30, 0, 0),
        ]
        self.updater.morahelper_mock.find_cut_dates.return_value = cut_dates

        self.updater._read_engagement = lambda user_uuid, date: engagements

        self.assertEqual(self.updater.recalculate_user("user_uuid"), {"user_uuid": 0})
        self.updater._ensure_primary.assert_has_calls(
            [
                call(
                    engagements[0],
                    "non_primary_uuid",
                    {"from": "2930-01-01", "to": None},
                ),
                call(
                    engagements[1],
                    "fixed_primary_uuid",
                    {"from": "2930-01-01", "to": None},
                ),
            ]
        )
        self.updater.morahelper_mock._mo_post.assert_not_called()



import asyncio
from contextlib import contextmanager
from functools import partial

from fastapi import FastAPI
from prometheus_fastapi_instrumentator import Instrumentator
from prometheus_client import Enum
from integrations.rundb.db_overview import DBOverview

from .config import get_changed_at_settings
from .sd_changed_at import changed_at

state = Enum(
    "sd_changed_at_state",
    "Current state of the SD Changed At integration",
    states=["running", "ok", "failure", "unknown"],
)
state.state("unknown")

app = FastAPI()
Instrumentator().instrument(app).expose(app)


@contextmanager
def update_state_metric() -> None:
    """Update SDChangedAt state metrics."""
    settings = get_changed_at_settings()
    run_db = settings.sd_import_run_db
    db_overview = DBOverview(run_db)
    try:
        status = db_overview._read_last_line("status")
        state.state("running")
        yield
        if "Running since" in status:
            state.state("failure")
        elif "Update finished" in status:
            state.state("ok")
        else:
            state.state("unknown")
    except:
        state.state("failure")
        raise


@app.get("/")
async def index() -> dict[str, str]:
    return {"name": "sdlon"}


@app.post("/trigger")
@app.get("/trigger", deprecated=True)
async def trigger(force: bool = False) -> dict[str, str]:
    loop = asyncio.get_running_loop()
    loop.call_soon(partial(update_state_metric()(changed_at), init=False, force=force))
    return {"triggered": "OK"}



from contextlib import asynccontextmanager
from typing import Iterable, Union

from more_itertools import bucket

from os2mo_data_import.Clients.LoRa.client import Client as LoRaClient
from os2mo_data_import.Clients.LoRa.model_parts.interface import LoraObj
from os2mo_data_import.Clients.MO.client import Client as MoClient
from os2mo_data_import.Clients.MO.model_parts.interface import MoObj


class MoLoRaClient:
    def __init__(self):
        self.__mo_client = MoClient()
        self.__lora_client = LoRaClient()

    @asynccontextmanager
    async def context(self):
        try:
            async with self.__mo_client.context():
                async with self.__lora_client.context():
                    yield
        finally:
            pass

    async def load_objs(
        self,
        objs: Iterable[Union[MoObj, LoraObj]],
        disable_progressbar: bool = False,
    ):
        mos = []
        loras = []
        for obj in objs:
            if isinstance(obj, MoObj):
                mos.append(obj)
            elif isinstance(obj, LoraObj):
                loras.append(obj)
            else:
                raise TypeError(f"unexpected type: {type(obj)}")

        await self.__mo_client.load_mo_objs(
            mos, disable_progressbar=disable_progressbar
        )
        await self.__lora_client.load_lora_objs(
            loras, disable_progressbar=disable_progressbar
        )



import datetime
import json
import logging
import os
import pickle
import re
import time
import urllib
from collections import defaultdict
from itertools import starmap
from operator import itemgetter
from pathlib import Path
from typing import Dict
from typing import Optional
from typing import Set
from typing import Tuple
from uuid import UUID

import click
import lora_utils
import requests
from dateutil import parser
from dateutil import tz
from more_itertools import bucket
from os2mo_dar_client import DARClient
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.load_settings import load_settings
from retrying import retry
from tqdm import tqdm

logger = logging.getLogger("LoraCache")

DEFAULT_TIMEZONE = tz.gettz("Europe/Copenhagen")

PICKLE_PROTOCOL = pickle.DEFAULT_PROTOCOL

LOG_LEVEL = logging.DEBUG
LOG_FILE = "lora_cache.log"


def get_rel_uuid_or_none(uuid, rel, item_name) -> Optional[str]:
    """Read uuid from rel. Log if it doesn't exist"""
    try:
        return rel[item_name][0]["uuid"]
    except IndexError:
        logger.error(f"Empty rel['{item_name}'] ({uuid=}), was {rel}")
    except KeyError:
        logger.info(f"No {item_name} found for {uuid=}, was {rel}")
    return None


class LoraCache:
    def __init__(self, resolve_dar=True, full_history=False, skip_past=False):
        msg = "Start LoRa cache, resolve dar: {}, full_history: {}"
        logger.info(msg.format(resolve_dar, full_history))
        self.resolve_dar = resolve_dar

        self.settings = self._load_settings()

        self.additional = {"relationer": ("tilknyttedeorganisationer", "tilhoerer")}

        self.dar_map = defaultdict(list)

        self.full_history = full_history
        self.skip_past = skip_past
        self.org_uuid = self._read_org_uuid()

    def _load_settings(self):
        return load_settings()

    def _read_org_uuid(self):
        mh = MoraHelper(hostname=self.settings["mora.base"], export_ansi=False)
        for attempt in range(0, 10):
            try:
                org_uuid = mh.read_organisation()
                return org_uuid
            except (
                json.decoder.JSONDecodeError,
                requests.exceptions.RequestException,
            ) as e:
                logger.error(e)
                print(e)
                time.sleep(5)
                continue
        # Unable to read org_uuid, must abort
        exit()

    def _get_effects(self, lora_object, relevant):
        effects = lora_utils.get_effects(
            lora_object["registreringer"][0],
            relevant=relevant,
            additional=self.additional,
        )
        # Notice, the code below will return the entire validity of an object
        # in the case of non-historic export, this could be handy in some
        # situations, eg ad->mo sync
        # if self.full_history:
        #     effects = lora_utils.get_effects(lora_object['registreringer'][0],
        #                                      relevant=relevant,
        #                                      additional=self.additional)
        # else:
        #     effects = lora_utils.get_effects(lora_object['registreringer'][0],
        #                                      relevant=self.additional,
        #                                      additional=relevant)
        return effects

    def _from_to_from_effect(self, effect):
        """
        Finds to and from date from an effect-row as returned by  iterating over the
        result of _get_effects().
        :param effect: The effect to analyse.
        :return: from_date and to_date. To date can be None, which should be
        interpreted as an infinite validity. In non-historic exports, both values
        can be None, meaning that this row is not the actual-state value.
        >>> lc=LoraCache(resolve_dar=False)
        >>> lc._from_to_from_effect([datetime.datetime(2020, 1, 1), datetime.datetime(2022, 1, 2)])
        ('2020-01-01', '2022-01-01')
        """

        dt_from = parser.isoparse(str(effect[0]))
        dt_from = dt_from.astimezone(DEFAULT_TIMEZONE)
        from_date = dt_from.date().isoformat()

        if effect[1].replace(tzinfo=None) == datetime.datetime.max:
            to_date = None
        else:
            dt_to = parser.isoparse(str(effect[1]))
            dt_to = dt_to.astimezone(DEFAULT_TIMEZONE)
            # MO considers end-dates inclusive, we need to subtract a day
            to_date = (dt_to.date() - datetime.timedelta(days=1)).isoformat()

        now = datetime.datetime.now(DEFAULT_TIMEZONE)
        # If this is an actual state export, we should only return a value if
        # the row is valid today.
        if not self.full_history:
            if to_date is None:
                # In this case, make sure dt_to is bigger than now
                dt_to = now + datetime.timedelta(days=1)
            if not dt_from <= now <= dt_to:
                from_date = to_date = None

        if self.skip_past:
            if to_date is None:
                # In this case, make sure dt_to is bigger than now
                dt_to = now + datetime.timedelta(days=1)
            if dt_to < now:
                from_date = to_date = None
        return from_date, to_date

    @retry(stop_max_attempt_number=7)
    def _perform_lora_lookup(self, url, params, skip_history=False, unit="it"):
        """
        Exctract a complete set of objects in LoRa.
        :param url: The url that should be used to extract data.
        :param skip_history: Force a validity of today, even if self.full_history
        is true.
        """
        # Copy to avoid modifying on retry
        params = params.copy()

        t = time.time()
        logger.debug("Start reading {}, params: {}, at t={}".format(url, params, t))
        results_pr_request = 1000
        params["foersteresultat"] = 0
        params["konsolider"] = True

        # Default, this can be overwritten in the lines below
        now = datetime.datetime.today()
        params["virkningFra"] = now.strftime("%Y-%m-%d") + " 00:00:00"
        params["virkningTil"] = now.strftime("%Y-%m-%d") + " 00:00:01"
        if self.full_history and not skip_history:
            params["virkningTil"] = "infinity"
            if not self.skip_past:
                params["virkningFra"] = "-infinity"

        response = requests.get(self.settings["mox.base"] + url, params=params)
        data = response.json()
        total = len(data["results"][0])

        params["list"] = 1
        params["maximalantalresultater"] = results_pr_request

        complete_data = []

        with tqdm(total=total, desc="Fetching " + unit, unit=unit) as pbar:
            while True:
                response = requests.get(self.settings["mox.base"] + url, params=params)
                data = response.json()
                results = data["results"]
                data_list = []
                if results:
                    data_list = data["results"][0]
                pbar.update(len(data_list))
                complete_data = complete_data + data_list
                if len(data_list) == 0:
                    break
                params["foersteresultat"] += results_pr_request
                logger.debug(
                    "Mellemtid, {} læsninger: {}s".format(
                        params["foersteresultat"], time.time() - t
                    )
                )
        logger.debug(
            "LoRa læsning færdig. {} elementer, {}s".format(
                len(complete_data), time.time() - t
            )
        )
        return complete_data

    def _cache_lora_facets(self):
        # Facets are eternal i MO and does not need a historic dump
        params = {"bvn": "%"}
        url = "/klassifikation/facet"
        facet_list = self._perform_lora_lookup(
            url, params, skip_history=True, unit="facet"
        )

        facets = {}
        for facet in tqdm(facet_list, desc="Processing facet", unit="facet"):
            uuid = facet["id"]
            reg = facet["registreringer"][0]
            user_key = reg["attributter"]["facetegenskaber"][0]["brugervendtnoegle"]
            facets[uuid] = {
                "user_key": user_key,
            }
        return facets

    def _cache_lora_classes(self):
        # MO itself will not read historic information on classes,
        # currently we replicate this behaviour here.
        params = {"bvn": "%"}
        url = "/klassifikation/klasse"
        class_list = self._perform_lora_lookup(
            url, params, skip_history=True, unit="class"
        )

        classes = {}
        for oio_class in tqdm(class_list, desc="Processing class", unit="class"):
            uuid = oio_class["id"]
            reg = oio_class["registreringer"][0]
            user_key = reg["attributter"]["klasseegenskaber"][0]["brugervendtnoegle"]
            scope = reg["attributter"]["klasseegenskaber"][0].get("omfang")
            title = reg["attributter"]["klasseegenskaber"][0]["titel"]
            facet = reg["relationer"]["facet"][0]["uuid"]
            classes[uuid] = {
                "user_key": user_key,
                "title": title,
                "scope": scope,
                "facet": facet,
            }
        return classes

    def _cache_lora_itsystems(self):
        # IT-systems are eternal i MO and does not need a historic dump
        params = {"bvn": "%"}
        url = "/organisation/itsystem"
        itsystem_list = self._perform_lora_lookup(
            url, params, skip_history=True, unit="itsystem"
        )

        itsystems = {}
        for itsystem in tqdm(
            itsystem_list, desc="Processing itsystem", unit="itsystem"
        ):
            uuid = itsystem["id"]
            reg = itsystem["registreringer"][0]
            user_key = reg["attributter"]["itsystemegenskaber"][0]["brugervendtnoegle"]
            name = reg["attributter"]["itsystemegenskaber"][0]["itsystemnavn"]

            itsystems[uuid] = {
                "user_key": user_key,
                "name": name,
            }
        return itsystems

    def _cache_lora_users(self):
        params = {"bvn": "%"}
        url = "/organisation/bruger"
        user_list = self._perform_lora_lookup(url, params, unit="user")

        relevant = {
            "attributter": ("brugeregenskaber", "brugerudvidelser"),
            "relationer": ("tilknyttedepersoner", "tilhoerer"),
            "tilstande": ("brugergyldighed",),
        }

        users = {}
        for user in tqdm(user_list, desc="Processing user", unit="user"):
            uuid = user["id"]
            users[uuid] = []

            effects = list(self._get_effects(user, relevant))
            for effect in effects:
                from_date, to_date = self._from_to_from_effect(effect)
                if from_date is None and to_date is None:
                    continue
                reg = effect[2]

                cpr = None
                tilknyttedepersoner = reg["relationer"].get("tilknyttedepersoner", [])
                if len(tilknyttedepersoner) == 0:
                    logger.warning("unable to find CPR for LoRa user %r", uuid)

                else:
                    cpr = tilknyttedepersoner[0]["urn"][-10:]

                egenskaber = reg["attributter"]["brugeregenskaber"]
                if len(egenskaber) == 0:
                    continue
                egenskaber = egenskaber[0]

                udv = reg["attributter"]["brugerudvidelser"]
                if len(udv) == 0:
                    continue
                udv = udv[0]

                user_key = egenskaber.get("brugervendtnoegle", "")
                fornavn = udv.get("fornavn", "")
                efternavn = udv.get("efternavn", "")
                kaldenavn_fornavn = udv.get("kaldenavn_fornavn", "")
                kaldenavn_efternavn = udv.get("kaldenavn_efternavn", "")
                users[uuid].append(
                    {
                        "uuid": uuid,
                        "cpr": cpr,
                        "user_key": user_key,
                        "fornavn": fornavn,
                        "efternavn": efternavn,
                        "navn": " ".join([fornavn, efternavn]).strip(),
                        "kaldenavn_fornavn": kaldenavn_fornavn,
                        "kaldenavn_efternavn": kaldenavn_efternavn,
                        "kaldenavn": " ".join(
                            [kaldenavn_fornavn, kaldenavn_efternavn]
                        ).strip(),
                        "from_date": from_date,
                        "to_date": to_date,
                    }
                )
        return users

    def _cache_lora_units(self):
        params = {"bvn": "%"}
        skip_history = False
        if not self.full_history:
            params["gyldighed"] = "Aktiv"
            skip_history = True
        url = "/organisation/organisationenhed"
        relevant = {
            "relationer": ("overordnet", "enhedstype", "niveau"),
            "attributter": ("organisationenhedegenskaber",),
        }

        unit_list = self._perform_lora_lookup(
            url, params, skip_history=skip_history, unit="unit"
        )

        units = {}
        for unit in tqdm(unit_list, desc="Processing unit", unit="unit"):
            uuid = unit["id"]
            units[uuid] = []

            effects = self._get_effects(unit, relevant)
            for effect in effects:
                from_date, to_date = self._from_to_from_effect(effect)
                if from_date is None and to_date is None:
                    continue
                relationer = effect[2]["relationer"]

                orgegenskaber = effect[2]["attributter"]["organisationenhedegenskaber"]
                if len(orgegenskaber) == 0:
                    continue
                egenskaber = orgegenskaber[0]
                parent_raw = relationer["overordnet"][0]["uuid"]
                if parent_raw == self.org_uuid:
                    parent = None
                else:
                    parent = parent_raw

                if "niveau" in relationer and len(relationer["niveau"]) > 0:
                    level = relationer["niveau"][0]["uuid"]
                else:
                    level = None
                units[uuid].append(
                    {
                        "uuid": uuid,
                        "user_key": egenskaber["brugervendtnoegle"],
                        "name": egenskaber["enhedsnavn"],
                        "unit_type": relationer["enhedstype"][0]["uuid"],
                        "level": level,
                        "parent": parent,
                        "from_date": from_date,
                        "to_date": to_date,
                    }
                )
        return units

    def _cache_lora_address(self):
        params = {"gyldighed": "Aktiv", "funktionsnavn": "Adresse"}

        url = "/organisation/organisationfunktion"
        relevant = {
            "relationer": (
                "tilknyttedeenheder",
                "tilknyttedebrugere",
                "adresser",
                "organisatoriskfunktionstype",
                "opgaver",
            ),
            "attributter": ("organisationfunktionegenskaber",),
        }
        address_list = self._perform_lora_lookup(url, params, unit="address")

        addresses = {}
        for address in tqdm(address_list, desc="Processing address", unit="address"):
            uuid = address["id"]
            addresses[uuid] = []

            effects = self._get_effects(address, relevant)
            for effect in effects:
                from_date, to_date = self._from_to_from_effect(effect)
                if from_date is None and to_date is None:
                    continue
                relationer = effect[2]["relationer"]

                if (
                    "tilknyttedeenheder" in relationer
                    and len(relationer["tilknyttedeenheder"]) > 0
                ):
                    unit_uuid = relationer["tilknyttedeenheder"][0]["uuid"]
                    user_uuid = None
                elif (
                    "tilknyttedebrugere" in relationer
                    and len(relationer["tilknyttedebrugere"]) > 0
                ):
                    user_uuid = relationer["tilknyttedebrugere"][0]["uuid"]
                    unit_uuid = None
                else:
                    # Skip if address is not attached to anything
                    continue

                dar_uuid = None
                value_raw = relationer["adresser"][0]["urn"]
                address_type = relationer["adresser"][0]["objekttype"]
                if address_type == "EMAIL":
                    scope = "E-mail"
                    skip_len = len("urn:mailto:")
                    value = value_raw[skip_len:]
                elif address_type == "WWW":
                    scope = "Url"
                    skip_len = len("urn:magenta.dk:www:")
                    value = value_raw[skip_len:]
                elif address_type == "PHONE":
                    scope = "Telefon"
                    skip_len = len("urn:magenta.dk:telefon:")
                    value = value_raw[skip_len:]
                elif address_type == "PNUMBER":
                    scope = "P-nummer"
                    skip_len = len("urn:dk:cvr:produktionsenhed:")
                    value = value_raw[skip_len:]
                elif address_type == "EAN":
                    scope = "EAN"
                    skip_len = len("urn:magenta.dk:ean:")
                    value = value_raw[skip_len:]
                elif address_type == "TEXT":
                    scope = "Text"
                    skip_len = len("urn:text:")
                    value = urllib.parse.unquote(value_raw[skip_len:])
                elif address_type == "MULTIFIELD_TEXT":
                    # This address type has more than one field
                    value_raw1 = relationer["adresser"][1]["urn"]
                    scope = "Multifield_text"
                    r1 = re.compile("urn:multifield_text:(.*)")
                    r2 = re.compile("urn:multifield_text2:(.*)")
                    # Ensure correct order so that "text" is before "text2"
                    value1 = r1.match(value_raw) or r1.match(value_raw1)
                    value2 = r2.match(value_raw) or r2.match(value_raw1)
                    # Both fields are put into one field in loracache as they are shown in MO
                    value = f"{value1.group(1)} :: {value2.group(1)}"
                    value = urllib.parse.unquote(value)
                elif address_type == "DAR":
                    scope = "DAR"
                    skip_len = len("urn:dar:")
                    dar_uuid = value_raw[skip_len:]
                    value = None

                    if self.dar_map is not None:
                        self.dar_map[dar_uuid].append(uuid)
                else:
                    print("Ny type: {}".format(address_type))
                    msg = "Unknown addresse type: {}, value: {}"
                    logger.error(msg.format(address_type, value_raw))
                    raise ("Unknown address type: {}".format(address_type))

                address_type_class = relationer["organisatoriskfunktionstype"][0][
                    "uuid"
                ]

                synlighed = None
                if relationer.get("opgaver"):
                    if relationer["opgaver"][0]["objekttype"] == "synlighed":
                        synlighed = relationer["opgaver"][0]["uuid"]

                addresses[uuid].append(
                    {
                        "uuid": uuid,
                        "user": user_uuid,
                        "unit": unit_uuid,
                        "value": value,
                        "scope": scope,
                        "dar_uuid": dar_uuid,
                        "adresse_type": address_type_class,
                        "visibility": synlighed,
                        "from_date": from_date,
                        "to_date": to_date,
                    }
                )
        return addresses

    def _cache_lora_engagements(self):
        params = {"gyldighed": "Aktiv", "funktionsnavn": "Engagement"}
        relevant = {
            "relationer": (
                "opgaver",
                "tilknyttedeenheder",
                "tilknyttedebrugere",
                "organisatoriskfunktionstype",
                "primær",
            ),
            "attributter": (
                "organisationfunktionegenskaber",
                "organisationfunktionudvidelser",
            ),
            "tilstande": ("organisationfunktiongyldighed",),
        }
        url = "/organisation/organisationfunktion"
        engagements = {}
        engagement_list = self._perform_lora_lookup(url, params, unit="engagement")
        for engagement in tqdm(
            engagement_list, desc="Processing engagement", unit="engagement"
        ):
            uuid = engagement["id"]

            effects = self._get_effects(engagement, relevant)
            engagement_effects = []
            for effect in effects:
                from_date, to_date = self._from_to_from_effect(effect)
                if from_date is None and to_date is None:
                    continue

                # Todo, this should be consistently implemented for all objects
                gyldighed = effect[2]["tilstande"]["organisationfunktiongyldighed"]
                if not gyldighed:
                    continue
                if not gyldighed[0]["gyldighed"] == "Aktiv":
                    continue

                attr = effect[2]["attributter"]
                rel = effect[2]["relationer"]

                if not rel["organisatoriskfunktionstype"]:
                    msg = "Missing in organisatoriskfunktionstype in {}"
                    logger.error(msg.format(engagement))
                    continue

                user_key = attr["organisationfunktionegenskaber"][0][
                    "brugervendtnoegle"
                ]

                engagement_type = rel["organisatoriskfunktionstype"][0]["uuid"]

                primary_type = None
                primær = rel.get("primær")
                if primær:
                    primary_type = primær[0]["uuid"]

                try:
                    job_function = rel["opgaver"][0]["uuid"]
                except Exception:
                    continue

                user_uuid = rel["tilknyttedebrugere"][0]["uuid"]
                unit_uuid = rel["tilknyttedeenheder"][0]["uuid"]

                udvidelser = {}
                udv_raw = attr.get("organisationfunktionudvidelser")
                if isinstance(udv_raw, list):
                    if len(udv_raw) == 1:
                        udvidelser = udv_raw[0]
                    if len(udv_raw) > 1:
                        msg = "Ugyldig organisationfunktionudvidelser: {}"
                        raise Exception(msg.format(udv_raw))
                fraction = udvidelser.get("fraktion")
                extensions = {
                    "udvidelse_1": udvidelser.get("udvidelse_1"),
                    "udvidelse_2": udvidelser.get("udvidelse_2"),
                    "udvidelse_3": udvidelser.get("udvidelse_3"),
                    "udvidelse_4": udvidelser.get("udvidelse_4"),
                    "udvidelse_5": udvidelser.get("udvidelse_5"),
                    "udvidelse_6": udvidelser.get("udvidelse_6"),
                    "udvidelse_7": udvidelser.get("udvidelse_7"),
                    "udvidelse_8": udvidelser.get("udvidelse_8"),
                    "udvidelse_9": udvidelser.get("udvidelse_9"),
                    "udvidelse_10": udvidelser.get("udvidelse_10"),
                }

                engagement_effects.append(
                    {
                        "uuid": uuid,
                        "user": user_uuid,
                        "unit": unit_uuid,
                        "fraction": fraction,
                        "user_key": user_key,
                        "engagement_type": engagement_type,
                        "primary_type": primary_type,
                        "job_function": job_function,
                        "extensions": extensions,
                        "from_date": from_date,
                        "to_date": to_date,
                    }
                )
            if engagement_effects:
                engagements[uuid] = engagement_effects
        return engagements

    def _cache_lora_associations(self):
        params = {"gyldighed": "Aktiv", "funktionsnavn": "Tilknytning"}
        relevant = {
            "relationer": (
                "tilknyttedeenheder",
                "tilknyttedebrugere",
                "organisatoriskfunktionstype",
                "tilknyttedefunktioner",
                "tilknyttedeitsystemer",
                "primær",
            ),
            "attributter": ("organisationfunktionegenskaber",),
        }
        url = "/organisation/organisationfunktion"
        associations = {}
        association_list = self._perform_lora_lookup(url, params, unit="association")
        for association in tqdm(
            association_list, desc="Processing association", unit="association"
        ):
            uuid = association["id"]
            associations[uuid] = []

            effects = self._get_effects(association, relevant)
            for effect in effects:
                from_date, to_date = self._from_to_from_effect(effect)
                if from_date is None and to_date is None:
                    continue

                attr = effect[2]["attributter"]
                rel = effect[2]["relationer"]

                if rel["tilknyttedeenheder"]:
                    unit_uuid = rel["tilknyttedeenheder"][0]["uuid"]
                else:
                    unit_uuid = None
                    logger.error("Error: Unable to find unit in {}".format(uuid))

                user_key = attr["organisationfunktionegenskaber"][0][
                    "brugervendtnoegle"
                ]
                association_type_uuid = get_rel_uuid_or_none(
                    uuid, rel, "organisatoriskfunktionstype"
                )
                user_uuid = get_rel_uuid_or_none(uuid, rel, "tilknyttedebrugere")
                it_user_uuid = get_rel_uuid_or_none(uuid, rel, "tilknyttedeitsystemer")
                job_function_uuid = (
                    get_rel_uuid_or_none(uuid, rel, "tilknyttedefunktioner")
                    if it_user_uuid
                    else None
                )

                primary_boolean = None
                primary_class_uuid = get_rel_uuid_or_none(uuid, rel, "primær")
                primary_class = self.classes.get(primary_class_uuid)
                if primary_class:
                    primary_boolean = primary_class.get("user_key", "") == "primary"

                associations[uuid].append(
                    {
                        "uuid": uuid,
                        "user": user_uuid,
                        "unit": unit_uuid,
                        "user_key": user_key,
                        "association_type": association_type_uuid,
                        "it_user": it_user_uuid,
                        "job_function": job_function_uuid,
                        "from_date": from_date,
                        "to_date": to_date,
                        "primary_boolean": primary_boolean,
                    }
                )

        return associations

    def _cache_lora_roles(self):
        params = {"gyldighed": "Aktiv", "funktionsnavn": "Rolle"}
        relevant = {
            "relationer": (
                "tilknyttedeenheder",
                "tilknyttedebrugere",
                "organisatoriskfunktionstype",
            )
        }
        url = "/organisation/organisationfunktion"
        roles = {}
        role_list = self._perform_lora_lookup(url, params, unit="role")
        for role in tqdm(role_list, desc="Processing role", unit="role"):
            uuid = role["id"]
            roles[uuid] = []

            effects = self._get_effects(role, relevant)
            for effect in effects:
                from_date, to_date = self._from_to_from_effect(effect)
                if from_date is None and to_date is None:
                    continue
                rel = effect[2]["relationer"]
                role_type = rel["organisatoriskfunktionstype"][0]["uuid"]
                user_uuid = rel["tilknyttedebrugere"][0]["uuid"]
                unit_uuid = rel["tilknyttedeenheder"][0]["uuid"]

                roles[uuid].append(
                    {
                        "uuid": uuid,
                        "user": user_uuid,
                        "unit": unit_uuid,
                        "role_type": role_type,
                        "from_date": from_date,
                        "to_date": to_date,
                    }
                )
        return roles

    def _cache_lora_leaves(self):
        params = {"gyldighed": "Aktiv", "funktionsnavn": "Orlov"}
        relevant = {
            "relationer": (
                "tilknyttedebrugere",
                "organisatoriskfunktionstype",
                "tilknyttedefunktioner",
            ),
            "attributter": ("organisationfunktionegenskaber",),
        }
        url = "/organisation/organisationfunktion"
        leaves = {}
        leave_list = self._perform_lora_lookup(url, params, unit="leave")
        for leave in tqdm(leave_list, desc="Processing leave", unit="leave"):
            uuid = leave["id"]
            leaves[uuid] = []
            effects = self._get_effects(leave, relevant)
            for effect in effects:
                from_date, to_date = self._from_to_from_effect(effect)
                if from_date is None and to_date is None:
                    continue
                attr = effect[2]["attributter"]
                rel = effect[2]["relationer"]
                user_key = attr["organisationfunktionegenskaber"][0][
                    "brugervendtnoegle"
                ]
                leave_type = rel["organisatoriskfunktionstype"][0]["uuid"]
                user_uuid = rel["tilknyttedebrugere"][0]["uuid"]

                if "tilknyttedefunktioner" in rel:
                    engagement_uuid = rel["tilknyttedefunktioner"][0]["uuid"]
                else:
                    engagement_uuid = None

                leaves[uuid].append(
                    {
                        "uuid": uuid,
                        "user": user_uuid,
                        "user_key": user_key,
                        "leave_type": leave_type,
                        "engagement": engagement_uuid,
                        "from_date": from_date,
                        "to_date": to_date,
                    }
                )
        return leaves

    def _cache_lora_it_connections(self):
        params = {"gyldighed": "Aktiv", "funktionsnavn": "IT-system"}
        url = "/organisation/organisationfunktion"
        it_connection_list = self._perform_lora_lookup(
            url, params, unit="it connection"
        )

        it_connections = {}
        for it_connection in tqdm(
            it_connection_list, desc="Processing it connection", unit="it connection"
        ):
            uuid = it_connection["id"]
            it_connections[uuid] = []

            relevant = {
                "relationer": (
                    "tilknyttedeenheder",
                    "tilknyttedebrugere",
                    "tilknyttedeitsystemer",
                    "primær",
                ),
                "attributter": ("organisationfunktionegenskaber",),
            }

            effects = self._get_effects(it_connection, relevant)
            for effect in effects:
                from_date, to_date = self._from_to_from_effect(effect)
                if from_date is None and to_date is None:
                    continue
                user_key = effect[2]["attributter"]["organisationfunktionegenskaber"][
                    0
                ]["brugervendtnoegle"]

                rel = effect[2]["relationer"]
                itsystem = rel["tilknyttedeitsystemer"][0]["uuid"]

                if "tilknyttedeenheder" in rel:
                    unit_uuid = rel["tilknyttedeenheder"][0]["uuid"]
                    user_uuid = None
                else:
                    user_uuid = rel["tilknyttedebrugere"][0]["uuid"]
                    unit_uuid = None

                primary_type = get_rel_uuid_or_none(uuid, rel, "primær")

                it_connections[uuid].append(
                    {
                        "uuid": uuid,
                        "user": user_uuid,
                        "unit": unit_uuid,
                        "username": user_key,
                        "itsystem": itsystem,
                        "primary_type": primary_type,
                        "from_date": from_date,
                        "to_date": to_date,
                    }
                )
        return it_connections

    def _cache_lora_kles(self):
        params = {"gyldighed": "Aktiv", "funktionsnavn": "KLE"}
        url = "/organisation/organisationfunktion"
        kle_list = self._perform_lora_lookup(url, params, unit="KLE")
        kles = {}
        for kle in tqdm(kle_list, desc="Processing KLE", unit="KLE"):
            uuid = kle["id"]
            kles[uuid] = []

            relevant = {
                "relationer": (
                    "opgaver",
                    "tilknyttedeenheder",
                    "organisatoriskfunktionstype",
                ),
                "attributter": ("organisationfunktionegenskaber",),
            }

            effects = self._get_effects(kle, relevant)
            for effect in effects:
                from_date, to_date = self._from_to_from_effect(effect)
                if from_date is None and to_date is None:
                    continue

                user_key = effect[2]["attributter"]["organisationfunktionegenskaber"][
                    0
                ]["brugervendtnoegle"]

                rel = effect[2]["relationer"]
                unit_uuid = rel["tilknyttedeenheder"][0]["uuid"]
                kle_number = rel["organisatoriskfunktionstype"][0]["uuid"]

                for aspekt in rel["opgaver"]:
                    kle_aspect = aspekt["uuid"]
                    kles[uuid].append(
                        {
                            "uuid": uuid,
                            "unit": unit_uuid,
                            "kle_number": kle_number,
                            "kle_aspect": kle_aspect,
                            "user_key": user_key,
                            "from_date": from_date,
                            "to_date": to_date,
                        }
                    )
        return kles

    def _cache_lora_related(self):
        params = {"gyldighed": "Aktiv", "funktionsnavn": "Relateret Enhed"}
        url = "/organisation/organisationfunktion"
        related_list = self._perform_lora_lookup(url, params, unit="related")
        related = {}
        for relate in tqdm(related_list, desc="Processing related", unit="related"):
            uuid = relate["id"]
            related[uuid] = []

            relevant = {"relationer": ("tilknyttedeenheder",), "attributter": ()}

            effects = self._get_effects(relate, relevant)
            for effect in effects:
                from_date, to_date = self._from_to_from_effect(effect)
                if from_date is None and to_date is None:
                    continue

                rel = effect[2]["relationer"]
                unit1_uuid = rel["tilknyttedeenheder"][0]["uuid"]
                unit2_uuid = rel["tilknyttedeenheder"][1]["uuid"]

                related[uuid].append(
                    {
                        "uuid": uuid,
                        "unit1_uuid": unit1_uuid,
                        "unit2_uuid": unit2_uuid,
                        "from_date": from_date,
                        "to_date": to_date,
                    }
                )
        return related

    def _cache_lora_managers(self):
        params = {"gyldighed": "Aktiv", "funktionsnavn": "Leder"}
        url = "/organisation/organisationfunktion"
        manager_list = self._perform_lora_lookup(url, params, unit="manager")

        managers = {}
        for manager in tqdm(manager_list, desc="Processing manager", unit="manager"):
            uuid = manager["id"]
            managers[uuid] = []
            relevant = {
                "relationer": (
                    "opgaver",
                    "tilknyttedeenheder",
                    "tilknyttedebrugere",
                    "organisatoriskfunktionstype",
                )
            }

            if self.full_history:
                effects = lora_utils.get_effects(
                    manager["registreringer"][0],
                    relevant=relevant,
                    additional=self.additional,
                )
            else:
                # Yes, `relevant` and `additional` *appear to be* switched
                # around by mistake in this method call! However, the code
                # seems to be deliberate - the data in `effects` cannot be
                # processed by `calculate_derived_unit_data` if the code is
                # changed to `relevant=relevant, additional=self.additional`.
                # TODO: Add unittest which captures this surprising behavior.
                effects = lora_utils.get_effects(
                    manager["registreringer"][0],
                    relevant=self.additional,
                    additional=relevant,
                )

            for effect in effects:
                from_date, to_date = self._from_to_from_effect(effect)
                rel = effect[2]["relationer"]

                user_uuid = get_rel_uuid_or_none(uuid, rel, "tilknyttedebrugere")
                unit_uuid = get_rel_uuid_or_none(uuid, rel, "tilknyttedeenheder")
                manager_type = get_rel_uuid_or_none(
                    uuid, rel, "organisatoriskfunktionstype"
                )

                manager_level = None  # populated in loop over "opgaver" below
                manager_responsibility = []

                for opgave in rel["opgaver"]:
                    if opgave["objekttype"] == "lederniveau":
                        manager_level = opgave["uuid"]
                    if opgave["objekttype"] == "lederansvar":
                        manager_responsibility.append(opgave["uuid"])

                managers[uuid].append(
                    {
                        "uuid": uuid,
                        "user": user_uuid,
                        "unit": unit_uuid,
                        "manager_type": manager_type,
                        "manager_level": manager_level,
                        "manager_responsibility": manager_responsibility,
                        "from_date": from_date,
                        "to_date": to_date,
                    }
                )

        return managers

    def calculate_primary_engagements(self):
        if self.full_history:
            msg = """
            Calculation of primary engagements is currently not implemented for
            full historic export.
            """
            print(msg)
            return

        def extract_engagement(uuid, eng_validities):
            """Extract engagement from engagement validities."""
            assert (len(eng_validities)) == 1
            eng = eng_validities[0]
            return uuid, eng

        def convert_to_scope_value(uuid, engagement):
            """Convert engagement to scope value."""
            primary_type = self.classes.get(engagement["primary_type"])
            if primary_type is None:
                logger.debug(
                    "Primary information missing in engagement {}".format(uuid)
                )
                return uuid, 0
            primary_scope = int(primary_type["scope"])
            return uuid, primary_scope

        def get_engagement_user(tup):
            uuid, engagement = tup
            return engagement["user"]

        # List of 2-tuples: uuid, engagement validities
        engagement_validities = self.engagements.items()
        # Iterator of 2-tuples: uuid, engagement
        engagements = starmap(extract_engagement, engagement_validities)
        # Buckets of iterators of 2-tuples: uuid, engagement
        user_buckets = bucket(engagements, key=get_engagement_user)
        # Run though the user buckets in turn
        for user_uuid in user_buckets:
            # Iterator of 2-tuples: uuid, engagement
            user_engagements = user_buckets[user_uuid]
            # Iterator of 2-tuples: uuid, scope_value
            scope_values = list(starmap(convert_to_scope_value, user_engagements))

            # Find the highest scope in the users engagements, all engagements with
            # this scope value will be considered primary.
            highest_scope = max(map(itemgetter(1), scope_values))

            # Loop through all engagements and initially set them as non-primary.
            for uuid, primary_scope in scope_values:
                self.engagements[uuid][0]["primary_boolean"] = False

            # Loop through all engagements and mark the first engagement of the
            # highest scope as the primary engagement
            for uuid, primary_scope in scope_values:
                if primary_scope == highest_scope:
                    logger.debug("Primary for {} is {}".format(user_uuid, uuid))
                    self.engagements[uuid][0]["primary_boolean"] = True
                    break

    def calculate_derived_unit_data(self):
        if self.full_history:
            msg = """
            Calculation of derived unit data is currently not implemented for
            full historic export.
            """
            print(msg)
            return

        responsibility_class = self.settings.get(
            "exporters.actual_state.manager_responsibility_class", None
        )
        for unit, unit_validities in self.units.items():
            assert (len(unit_validities)) == 1
            unit_info = unit_validities[0]
            manager_uuid = None
            acting_manager_uuid = None

            def find_manager_for_org_unit(org_unit):
                def to_manager_info(manager_uuid, manager_validities):
                    assert (len(manager_validities)) == 1
                    manager_info = manager_validities[0]
                    return manager_uuid, manager_info

                def filter_invalid_managers(tup):
                    manager_uuid, manager_info = tup
                    # Wrong unit
                    if manager_info["unit"] != org_unit:
                        return False
                    # No resonsibility class
                    if responsibility_class is None:
                        return True
                    # Check responsability
                    return any(
                        resp == responsibility_class
                        for resp in manager_info["manager_responsibility"]
                    )

                managers = self.managers.items()
                managers = starmap(to_manager_info, managers)
                managers = filter(filter_invalid_managers, managers)
                managers = map(itemgetter(0), managers)
                return next(managers, None)

            # Find a direct manager, if possible
            manager_uuid = acting_manager_uuid = find_manager_for_org_unit(unit)

            location = ""
            current_unit = unit_info
            while current_unit:
                location = current_unit["name"] + "\\" + location
                current_parent = current_unit.get("parent")
                if current_parent is not None and current_parent in self.units:
                    current_unit = self.units[current_parent][0]
                else:
                    current_unit = None

                # Find the acting manager.
                if acting_manager_uuid is None:
                    acting_manager_uuid = find_manager_for_org_unit(current_parent)
            location = location[:-1]

            self.units[unit][0]["location"] = location
            self.units[unit][0]["manager_uuid"] = manager_uuid
            self.units[unit][0]["acting_manager_uuid"] = acting_manager_uuid

    def _read_from_dar(self, uuids: Set[UUID]) -> Tuple[Dict, Set[UUID]]:
        with DARClient() as dc:
            return dc.fetch(uuids)

    def _cache_dar(self):
        # Initialize cache for entries we cannot lookup
        dar_uuids = missing = set(self.dar_map.keys())
        dar_cache = dict(
            map(lambda dar_uuid: (dar_uuid, {"betegnelse": None}), dar_uuids)
        )
        if self.resolve_dar:
            dar_hits, missing = self._read_from_dar(dar_uuids)
            # dar_hits is a dict with UUIDs as keys. We need to cast them to strings.
            dar_hits_uuids_as_str = map(str, dar_hits.keys())
            dar_hits = dict(zip(dar_hits_uuids_as_str, dar_hits.values()))
            dar_cache.update(dar_hits)
            logger.info(f"Total dar: {len(dar_uuids)}, no-hit: {len(missing)}")
            for dar_uuid, uuid_list in self.dar_map.items():
                for uuid in uuid_list:
                    for address in self.addresses[uuid]:
                        address["value"] = dar_cache[dar_uuid].get("betegnelse")
        logger.info(f"Total dar: {len(dar_uuids)}, no-hit: {len(missing)}")
        return dar_cache

    def populate_cache(self, dry_run=None, skip_associations=False):
        """
        Perform the actual data import.
        :param skip_associations: If associations are not needed, they can be
        skipped for increased performance.
        :param dry_run: For testing purposes it is possible to read from cache.
        """
        if dry_run is None:
            dry_run = os.environ.get("USE_CACHED_LORACACHE", False)

        # Ensure that tmp/ exists
        Path("tmp/").mkdir(exist_ok=True)

        if self.full_history:
            if self.skip_past:
                facets_file = "tmp/facets_historic_skip_past.p"
                classes_file = "tmp/classes_historic_skip_past.p"
                users_file = "tmp/users_historic_skip_past.p"
                units_file = "tmp/units_historic_skip_past.p"
                addresses_file = "tmp/addresses_historic_skip_past.p"
                engagements_file = "tmp/engagements_historic_skip_past.p"
                managers_file = "tmp/managers_historic_skip_past.p"
                associations_file = "tmp/associations_historic_skip_past.p"
                leaves_file = "tmp/leaves_historic_skip_past.p"
                roles_file = "tmp/roles_historic_skip_past.p"
                itsystems_file = "tmp/itsystems_historic_skip_past.p"
                it_connections_file = "tmp/it_connections_historic_skip_past.p"
                kles_file = "tmp/kles_historic_skip_past.p"
                related_file = "tmp/related_historic_skip_past.p"
            else:
                facets_file = "tmp/facets_historic.p"
                classes_file = "tmp/classes_historic.p"
                users_file = "tmp/users_historic.p"
                units_file = "tmp/units_historic.p"
                addresses_file = "tmp/addresses_historic.p"
                engagements_file = "tmp/engagements_historic.p"
                managers_file = "tmp/managers_historic.p"
                associations_file = "tmp/associations_historic.p"
                leaves_file = "tmp/leaves_historic.p"
                roles_file = "tmp/roles_historic.p"
                itsystems_file = "tmp/itsystems_historic.p"
                it_connections_file = "tmp/it_connections_historic.p"
                kles_file = "tmp/kles_historic.p"
                related_file = "tmp/related_historic.p"
        else:
            facets_file = "tmp/facets.p"
            classes_file = "tmp/classes.p"
            users_file = "tmp/users.p"
            units_file = "tmp/units.p"
            addresses_file = "tmp/addresses.p"
            engagements_file = "tmp/engagements.p"
            managers_file = "tmp/managers.p"
            associations_file = "tmp/associations.p"
            leaves_file = "tmp/leaves.p"
            roles_file = "tmp/roles.p"
            itsystems_file = "tmp/itsystems.p"
            it_connections_file = "tmp/it_connections.p"
            kles_file = "tmp/kles.p"
            related_file = "tmp/related.p"

        if dry_run:
            logger.info("LoRa cache dry run - no actual read")
            with open(facets_file, "rb") as f:
                self.facets = pickle.load(f)
            with open(classes_file, "rb") as f:
                self.classes = pickle.load(f)
            with open(users_file, "rb") as f:
                self.users = pickle.load(f)
            with open(units_file, "rb") as f:
                self.units = pickle.load(f)
            with open(addresses_file, "rb") as f:
                self.addresses = pickle.load(f)
            with open(engagements_file, "rb") as f:
                self.engagements = pickle.load(f)
            with open(managers_file, "rb") as f:
                self.managers = pickle.load(f)

            if not skip_associations:
                with open(associations_file, "rb") as f:
                    self.associations = pickle.load(f)

            with open(leaves_file, "rb") as f:
                self.leaves = pickle.load(f)
            with open(roles_file, "rb") as f:
                self.roles = pickle.load(f)
            with open(itsystems_file, "rb") as f:
                self.itsystems = pickle.load(f)
            with open(it_connections_file, "rb") as f:
                self.it_connections = pickle.load(f)
            with open(kles_file, "rb") as f:
                self.kles = pickle.load(f)
            with open(related_file, "rb") as f:
                self.related = pickle.load(f)
            self.dar_cache = {}
            return

        t = time.time()  # noqa: F841
        msg = "Kørselstid: {:.1f}s, {} elementer, {:.0f}/s"  # noqa: F841

        # Here we should activate read-only mode
        def read_facets():
            logger.info("Læs facetter")
            self.facets = self._cache_lora_facets()
            return self.facets

        def read_classes():
            logger.info("Læs klasser")
            self.classes = self._cache_lora_classes()
            return self.classes

        def read_users():
            logger.info("Læs brugere")
            self.users = self._cache_lora_users()
            return self.users

        def read_units():
            logger.info("Læs enheder")
            self.units = self._cache_lora_units()
            return self.units

        def read_addresses():
            logger.info("Læs adresser:")
            self.addresses = self._cache_lora_address()
            return self.addresses

        def read_engagements():
            logger.info("Læs engagementer")
            self.engagements = self._cache_lora_engagements()
            return self.engagements

        def read_managers():
            logger.info("Læs ledere")
            self.managers = self._cache_lora_managers()
            return self.managers

        def read_associations():
            logger.info("Læs tilknytninger")
            self.associations = self._cache_lora_associations()
            return self.associations

        def read_leaves():
            logger.info("Læs orlover")
            self.leaves = self._cache_lora_leaves()
            return self.leaves

        def read_roles():
            logger.info("Læs roller")
            self.roles = self._cache_lora_roles()
            return self.roles

        def read_itsystems():
            logger.info("Læs itsystem")
            self.itsystems = self._cache_lora_itsystems()
            return self.itsystems

        def read_it_connections():
            logger.info("Læs it kobling")
            self.it_connections = self._cache_lora_it_connections()
            return self.it_connections

        def read_kles():
            logger.info("Læs kles")
            self.kles = self._cache_lora_kles()
            return self.kles

        def read_related():
            logger.info("Læs enhedssammenkobling")
            self.related = self._cache_lora_related()
            return self.related

        def read_dar():
            logger.info("Læs dar")
            self.dar_cache = self._cache_dar()
            # with open(cache_file, 'wb') as f:
            #    pickle.dump(self.dar_cache, f, pickle.HIGHEST_PROTOCOL)

        tasks = []
        tasks.append((read_facets, facets_file))
        tasks.append((read_classes, classes_file))
        tasks.append((read_users, users_file))
        tasks.append((read_units, units_file))
        tasks.append((read_addresses, addresses_file))
        tasks.append((read_engagements, engagements_file))
        tasks.append((read_managers, managers_file))
        if not skip_associations:
            tasks.append((read_associations, associations_file))
        tasks.append((read_leaves, leaves_file))
        tasks.append((read_roles, roles_file))
        tasks.append((read_itsystems, itsystems_file))
        tasks.append((read_it_connections, it_connections_file))
        tasks.append((read_kles, kles_file))
        tasks.append((read_related, related_file))
        tasks.append((read_dar, None))

        for task, filename in tqdm(tasks, desc="LoraCache", unit="task"):
            data = task()
            if filename:
                with open(filename, "wb") as f:
                    pickle.dump(data, f, PICKLE_PROTOCOL)

        # Here we should de-activate read-only mode


def fetch_loracache() -> Tuple[LoraCache, LoraCache]:
    # Here we should activate read-only mode, actual state and
    # full history dumps needs to be in sync.

    # Full history does not calculate derived data, we must
    # fetch both kinds.
    lc = LoraCache(resolve_dar=True, full_history=False)
    lc.populate_cache(skip_associations=True)
    lc.calculate_derived_unit_data()
    lc.calculate_primary_engagements()

    # Todo, in principle it should be possible to run with skip_past True
    # This is now fixed in a different branch, remember to update when
    # merged.
    lc_historic = LoraCache(resolve_dar=True, full_history=True, skip_past=False)
    lc_historic.populate_cache(skip_associations=True)
    # Here we should de-activate read-only mode
    return lc, lc_historic


@click.command()
@click.option("--historic/--no-historic", default=True, help="Do full historic export")
@click.option(
    "--skip-past", is_flag=True, default=False, help="Skip past in historic export"
)
@click.option(
    "--resolve-dar/--no-resolve-dar", default=False, help="Resolve DAR addresses"
)
@click.option("--read-from-cache", is_flag=True)
def cli(historic, skip_past, resolve_dar, read_from_cache):
    lc = LoraCache(
        full_history=historic,
        skip_past=skip_past,
        resolve_dar=resolve_dar,
    )
    lc.populate_cache(dry_run=read_from_cache)

    logger.info("Now calcualate derived data")
    lc.calculate_derived_unit_data()
    lc.calculate_primary_engagements()


if __name__ == "__main__":

    for name in logging.root.manager.loggerDict:  # type: ignore
        if name in ("LoraCache"):
            logging.getLogger(name).setLevel(LOG_LEVEL)
        else:
            logging.getLogger(name).setLevel(logging.ERROR)

    logging.basicConfig(
        format="%(levelname)s %(asctime)s %(name)s %(message)s",
        level=LOG_LEVEL,
        filename=LOG_FILE,
    )

    cli()



from csv import DictWriter
from io import StringIO

import pydantic
from hypothesis import strategies as st


def text_except(*exceptions: str):
    """Return an `st.text` strategy where certain characters are not present."""
    return st.text(
        alphabet=st.characters(
            # Exclude our list of excepted characters
            blacklist_characters=exceptions,
            # Exclude surrogate characters because they are invalid in the
            # UTF-8 encoding. See:
            # https://hypothesis.readthedocs.io/en/latest/data.html#hypothesis.strategies.text
            blacklist_categories=("Cs",),
        )
    )


@st.composite
def csv_buf_from_model(draw, model: pydantic.BaseModel, delimiter: str = ";"):
    """Return a `StringIO` instance containing a single-item CSV built from
    the Pydantic model `model`.
    """
    # The `csv` module does not handle ASCII NUL characters well. Since we are
    # not testing the `csv` module itself, but rather the code utilizing it,
    # we'll ensure we do not produce sample CSV data containing ASCII NULs.
    # We'll also make sure to not include the CSV delimiter character in the
    # values produced for the CSV fields, as this will also cause problems when
    # constructing a mock CSV buffer.
    ascii_nul = chr(0)
    overrides = {
        fieldname: text_except(ascii_nul, delimiter)
        for fieldname, field in model.__fields__.items()
        if issubclass(field.type_, str)
    }

    # Construct a strategy based on the Pydantic model `model` and the contents
    # of `overrides`.
    strategy: st.SearchStrategy = st.builds(model, **overrides)  # type: ignore

    # Construct a buffer containing a one-item CSV file based on our `strategy`
    instance = draw(strategy)  # type: ignore
    buf = StringIO()
    csv_row = {
        getattr(field, "alias", key): getattr(instance, key)
        for key, field in instance.__fields__.items()
    }
    writer = DictWriter(buf, list(csv_row.keys()), delimiter=delimiter, escapechar="\\")
    writer.writeheader()
    writer.writerow(csv_row)
    return buf



import collections
import logging
from typing import Dict
from typing import Optional

import click
from glom import assign
from ra_utils.load_settings import load_settings


logger = logging.getLogger("AdReader")


def _read_global_settings(top_settings):
    global_settings = {}
    global_settings["mora.base"] = top_settings.get("mora.base")
    global_settings["servers"] = top_settings.get("integrations.ad")[0].get(
        "servers", []
    )
    global_settings["winrm_host"] = top_settings.get("integrations.ad.winrm_host")
    global_settings["system_user"] = top_settings["integrations.ad"][0]["system_user"]
    global_settings["password"] = top_settings["integrations.ad"][0]["password"]
    if not global_settings["winrm_host"]:
        msg = "Missing hostname for remote management server"
        logger.error(msg)
        raise Exception(msg)
    return global_settings


def _read_primary_ad_settings(top_settings, index=0):
    primary_settings = {}

    if top_settings.get("integrations.ad") is None:
        raise Exception("integration.ad settings not found")

    if len(top_settings["integrations.ad"]) < (index + 1):
        raise Exception("ad index %d not found" % index)

    # settings that must be in place
    primary_settings["search_base"] = top_settings["integrations.ad"][index].get(
        "search_base"
    )
    primary_settings["cpr_field"] = top_settings["integrations.ad"][index].get(
        "cpr_field"
    )
    primary_settings["system_user"] = top_settings["integrations.ad"][index].get(
        "system_user"
    )
    primary_settings["password"] = top_settings["integrations.ad"][index].get(
        "password"
    )
    primary_settings["properties"] = top_settings["integrations.ad"][index].get(
        "properties"
    )

    missing = []
    for key, val in primary_settings.items():
        if val is None:
            missing.append(key)

    # 36182 exclude non primary AD-users
    primary_settings["discriminator.field"] = top_settings["integrations.ad"][
        index
    ].get("discriminator.field")
    if primary_settings["discriminator.field"] is not None:

        # if we have a field we MUST have .values and .function
        primary_settings["discriminator.values"] = top_settings["integrations.ad"][
            index
        ].get("discriminator.values")
        if primary_settings["discriminator.values"] is None:
            missing.append("discriminator.values")

        primary_settings["discriminator.function"] = top_settings["integrations.ad"][
            index
        ].get("discriminator.function")
        if primary_settings["discriminator.function"] is None:
            missing.append("discriminator.function")

        if not primary_settings["discriminator.function"] in ["include", "exclude"]:
            raise ValueError(
                "'ad.discriminator.function'"
                + " must be 'include' or 'exclude' for AD %d" % index
            )

    # Settings that do not need to be set, or have defaults
    index_settings = top_settings["integrations.ad"][index]
    primary_settings["servers"] = index_settings.get("servers", [])
    primary_settings["caseless_samname"] = index_settings.get("caseless_samname", True)
    primary_settings["sam_filter"] = index_settings.get("sam_filter", "")
    primary_settings["cpr_separator"] = index_settings.get("cpr_separator", "")
    primary_settings["pseudo_cprs"] = index_settings.get("pseudo_cprs", [])

    primary_settings["method"] = index_settings.get("method", "kerberos")

    primary_settings["ad_mo_sync_mapping"] = index_settings.get(
        "ad_mo_sync_mapping", {}
    )
    primary_settings["ad_mo_sync_terminate_missing"] = index_settings.get(
        "ad_mo_sync_terminate_missing", False
    )
    primary_settings[
        "ad_mo_sync_terminate_missing_require_itsystem"
    ] = index_settings.get("ad_mo_sync_terminate_missing_require_itsystem", True)
    primary_settings["ad_mo_sync_terminate_disabled"] = index_settings.get(
        "ad_mo_sync_terminate_disabled"
    )
    primary_settings["ad_mo_sync_pre_filters"] = index_settings.get(
        "ad_mo_sync_pre_filters", []
    )
    primary_settings["ad_mo_sync_terminate_disabled_filters"] = index_settings.get(
        "ad_mo_sync_terminate_disabled_filters", []
    )

    if missing:
        msg = "Missing settings in AD {}: {}".format(index, missing)
        logger.error(msg)
        raise Exception(msg)

    return primary_settings


def _read_primary_write_information(top_settings):
    """
    Read the configuration for writing to the primary AD. If anything is missing,
    the AD write will be disabled.
    """
    # TODO: Some happy day, we could check for the actual validity of these

    # Straight-forward field mappings
    required_keys = {
        "uuid_field",
        "level2orgunit_field",
        "level2orgunit_type",
        "upn_end",
    }
    conf = {
        key: top_settings["integrations.ad.write.%s" % key] for key in required_keys
    }

    # Special field mappings
    conf["cpr_field"] = top_settings["integrations.ad"][0]["cpr_field"]
    conf["org_field"] = top_settings["integrations.ad.write.org_unit_field"]

    missing = {key for key, val in conf.items() if val is None}
    if missing:
        msg = "Missing values for AD write: %r"
        logger.info(msg, missing)
        raise ValueError(msg % missing)

    # Template fields
    conf["mo_to_ad_fields"] = top_settings.get(
        "integrations.ad_writer.mo_to_ad_fields", {}
    )
    conf["template_to_ad_fields"] = top_settings.get(
        "integrations.ad_writer.template_to_ad_fields", {}
    )

    # Overrides the "-Path" argument to "New-ADUser", if set
    if "integrations.ad_writer.new_ad_user_path" in top_settings:
        conf["new_ad_user_path"] = top_settings[
            "integrations.ad_writer.new_ad_user_path"
        ]

    # Check for illegal configuration of AD Write.
    mo_to_ad_fields = conf["mo_to_ad_fields"]
    template_to_ad_fields = conf["template_to_ad_fields"]
    other_ad_field_names = [conf["org_field"], conf["uuid_field"]]
    if conf.get("level2orgunit_field"):
        other_ad_field_names.append(conf["level2orgunit_field"])
    ad_field_names = (
        list(mo_to_ad_fields.values())
        + list(template_to_ad_fields.keys())
        + other_ad_field_names
    )
    # Conflicts are case-insensitive
    counter = collections.Counter(map(str.lower, ad_field_names))
    dupes = sorted(set(name for name, count in counter.items() if count > 1))
    if dupes:
        msg = "Duplicated AD field names in settings: %r"
        logger.info(msg, dupes)
        raise ValueError(msg % dupes)

    # Check that all settings we write to are in properties for all ADs
    for ad_settings in top_settings["integrations.ad"]:
        properties = set(map(str.lower, ad_settings.get("properties", [])))
        missing_properties = list(
            filter(
                lambda ad_field: ad_field != "" and ad_field.lower() not in properties,
                ad_field_names,
            )
        )
        if missing_properties:
            msg = "Missing AD field names in properties: %r"
            logger.info(msg, missing_properties)
            raise ValueError(msg % missing_properties)

    return conf


def read_settings(top_settings=None, index=0):
    if top_settings is None:
        top_settings = load_settings()

    settings = {}
    settings["global"] = _read_global_settings(top_settings)
    settings["primary"] = _read_primary_ad_settings(top_settings, index)
    # TODO: better check for AD-writer.
    if "integrations.ad.write.level2orgunit_field" in top_settings:
        settings["primary_write"] = _read_primary_write_information(top_settings)
    return settings


def injected_settings(
    settings_key: str,
    ad_settings: Optional[Dict] = None,
    normal_settings: Optional[Dict] = None,
) -> Dict:
    ad_settings = ad_settings or read_settings()
    normal_settings = normal_settings or load_settings()
    inject = normal_settings.get(settings_key, {})
    for path, value in inject.items():
        assign(ad_settings, path, value)
    return ad_settings


@click.command()
@click.option("--inject", is_flag=True, help="inject AD_life_cycle settings")
def show_ad_settings(inject):
    settings = (
        injected_settings("ad_lifecycle_injected_settings")
        if inject
        else read_settings()
    )
    click.echo(settings)


if __name__ == "__main__":
    show_ad_settings()



from typing import Any
from typing import Dict

from jinja2 import Environment
from jinja2 import StrictUndefined
from jinja2 import Template

from .utils import dict_map
from .utils import dict_partition
from .utils import duplicates
from .utils import lower_list


class InvalidValue:
    _value = "<invalid value>"

    def __str__(self):
        return self._value

    def __eq__(self, other):
        return self._value == other


INVALID = InvalidValue()


# Parameters that should not be quoted
no_quote_list = ["Credential", "Enabled", "AccountPassword"]


cmdlet_parameters = {
    # https://docs.microsoft.com/en-us/powershell/module/activedirectory/new-aduser
    "New-ADUser": {
        "AccountExpirationDate",
        "AccountNotDelegated",
        "AccountPassword",
        "AllowReversiblePasswordEncryption",
        "AuthType",
        "CannotChangePassword",
        "Certificates",
        "ChangePasswordAtLogon",
        "City",
        "Company",
        "CompoundIdentitySupported",
        "Country",
        "Credential",
        "Department",
        "Description",
        "DisplayName",
        "Division",
        "EmailAddress",
        "EmployeeID",
        "EmployeeNumber",
        "Enabled",
        "Fax",
        "GivenName",
        "HomeDirectory",
        "HomeDrive",
        "HomePage",
        "HomePhone",
        "Initials",
        "Instance",
        "KerberosEncryptionType",
        "LogonWorkstations",
        "Manager",
        "MobilePhone",
        "Name",
        "Office",
        "OfficePhone",
        "Organization",
        "OtherAttributes",
        "OtherName",
        "PasswordNeverExpires",
        "PasswordNotRequired",
        "Path",
        "POBox",
        "PostalCode",
        "PrincipalsAllowedToDelegateToAccount",
        "ProfilePath",
        "SamAccountName",
        "ScriptPath",
        "Server",
        "ServicePrincipalNames",
        "SmartcardLogonRequired",
        "State",
        "StreetAddress",
        "Surname",
        "Title",
        "TrustedForDelegation",
        "Type",
        "UserPrincipalName",
        # "WhatIf", "Confirm", "PassThru",
    },
    # https://docs.microsoft.com/en-us/powershell/module/activedirectory/set-aduser
    "Set-ADUser": {
        "AccountExpirationDate",
        "AccountNotDelegated",
        "Add",
        "AllowReversiblePasswordEncryption",
        "AuthType",
        "CannotChangePassword",
        "Certificates",
        "ChangePasswordAtLogon",
        "City",
        "Clear",
        "Company",
        "CompoundIdentitySupported",
        "Country",
        "Credential",
        "Department",
        "Description",
        "DisplayName",
        "Division",
        "EmailAddress",
        "EmployeeID",
        "EmployeeNumber",
        "Enabled",
        "Fax",
        "GivenName",
        "HomeDirectory",
        "HomeDrive",
        "HomePage",
        "HomePhone",
        "Identity",
        "Initials",
        "KerberosEncryptionType",
        "LogonWorkstations",
        "Manager",
        "MobilePhone",
        "Office",
        "OfficePhone",
        "Organization",
        "OtherName",
        "Partition",
        "PasswordNeverExpires",
        "PasswordNotRequired",
        "POBox",
        "PostalCode",
        "PrincipalsAllowedToDelegateToAccount",
        "ProfilePath",
        "Remove",
        "Replace",
        "SamAccountName",
        "ScriptPath",
        "Server",
        "ServicePrincipalNames",
        "SmartcardLogonRequired",
        "State",
        "StreetAddress",
        "Surname",
        "Title",
        "TrustedForDelegation",
        "UserPrincipalName",
        # "WhatIf", "Confirm", "PassThru",
    },
}

# These may never be emitted in parameters / other_attributes
illegal_parameters = {
    "New-ADUser": ["Manager"],
    "Set-ADUser": ["Manager", "Name"],
}
illegal_attributes = {
    "New-ADUser": ["Credential", "Name"],
    "Set-ADUser": ["Credential", "Name"],
}

cmdlet_templates = {
    "New-ADUser": """
        New-ADUser
        {%- for parameter, value in parameters.items() %}
          -{{ parameter }} {{ value }}
        {%- endfor %}
        {% if other_attributes %}
          -OtherAttributes @{
        {%- for attribute, value in other_attributes.items() -%}
            "{{ attribute }}"={{ value }};
        {%- endfor -%}
        }
        {% endif %}
    """,
    # Update information saved on a user
    # Notice: Name cannot be updated using Set-ADUser, this must be done
    # with Rename-AdObject
    # TODO: Consider Replace versus Remove/Clean/Add
    "Set-ADUser": """
        Get-ADUser
          -Filter 'SamAccountName -eq {{ parameters['SamAccountName'] }}'
          -Credential {{ parameters['Credential'] }} |
        Set-ADUser
        {%- for parameter, value in parameters.items() %}
          -{{ parameter }} {{ value }}
        {%- endfor %}
          -Replace @{
        {%- for attribute, value in other_attributes.items() -%}
            "{{ attribute }}"={{ value }};
        {%- endfor -%}
        }
    """,
}


def prepare_settings_based_field_templates(jinja_map, cmd, settings):
    """Expand jinja_map with settings based templates.

    Args:
        jinja_map: dictionary from ad field names to jinja template strings.
        cmd: command to generate template for.
        settings: dictionary containing settings from settings.json

    Returns:
        dict: A jinja_map which has been extended with settings based values.
    """
    # Build settings-based templates
    def _get_setting_type(settings, key):
        # TODO: Currently we ignore school
        try:
            result = settings[key]
        except KeyError:
            msg = "Unable to find settings type: " + key
            print(msg)
        if not result:
            raise Exception("%r is empty" % key)
        return result

    write_settings = _get_setting_type(settings, "primary_write")
    primary_settings = _get_setting_type(settings, "primary")

    if write_settings.get("level2orgunit_field"):
        jinja_map[
            write_settings["level2orgunit_field"]
        ] = "{{ mo_values['level2orgunit'] }}"

    jinja_map[write_settings["org_field"]] = "{{ mo_values['location'] }}"

    # Local fields for MO->AD sync'ing
    named_sync_fields = write_settings.get("mo_to_ad_fields")
    for mo_field, ad_field in named_sync_fields.items():
        jinja_map[ad_field] = "{{ mo_values['" + mo_field + "'] }}"

    # Local fields for MO->AD sync'ing
    named_sync_template_fields = write_settings.get("template_to_ad_fields")
    for ad_field, template in named_sync_template_fields.items():
        jinja_map[ad_field] = template

    if cmd == "New-ADUser":  # New user
        jinja_map["UserPrincipalName"] = "{{ user_sam }}@" + write_settings["upn_end"]
        jinja_map[write_settings["uuid_field"]] = "{{ mo_values['uuid'] }}"

        # If local settings dictates a separator, we add it directly to the
        # power-shell code.
        jinja_map[write_settings["cpr_field"]] = (
            "{{ mo_values['cpr'][0:6] }}"
            + primary_settings["cpr_separator"]
            + "{{ mo_values['cpr'][6:10] }}"
        )

    return jinja_map


def prepare_and_check_login_field_templates(jinja_map):
    """Check validity and expand jinja_map with login templates.

    Args:
        jinja_map: dictionary from ad field names to jinja template strings.

    Returns:
        dict: A jinja_map which has been extended with login templates.
    """
    # Check against hardcoded values, as these will be forcefully overridden.
    jinja_keys = lower_list(jinja_map.keys())
    if "credential" in jinja_keys:
        raise ValueError("Credential is hardcoded")
    if "samaccountname" in jinja_keys:
        raise ValueError("SamAccountName is hardcoded")
    if "manager" in jinja_keys:
        raise ValueError("Manager is handled uniquely")
    # Do the forceful override
    jinja_map["Credential"] = "$usercredential"
    jinja_map["SamAccountName"] = "{{ user_sam }}"

    return jinja_map


def prepare_field_templates(cmd, settings, jinja_map=None):
    """Build a finalized map of parameters and attributes.

    Args:
        cmd: command to generate template for.
        settings: dictionary containing settings from settings.json
        jinja_map: dictionary from ad field names to jinja template strings.

    Returns:
        tuple(dict, dict):
            parameters: a dict of parameter key, value pairs
            other_attributes: a dict of attribute key, value pairs

            Both dicts have the same format, namely:
                field_name -> jinja template for the field
    """
    # Load field templates (ad_field --> template)
    jinja_map = jinja_map or {}
    jinja_map = prepare_settings_based_field_templates(jinja_map, cmd, settings)
    jinja_map = prepare_and_check_login_field_templates(jinja_map)

    # Check against duplicates in jinja_map
    ad_fields_low = map(lambda ad_field: ad_field.lower(), jinja_map.keys())
    duplicate_ad_fields = duplicates(ad_fields_low)
    if duplicate_ad_fields:
        raise ValueError("Duplicate ad_field: " + ",".join(duplicate_ad_fields))
    return jinja_map


def quote_templates(jinja_map):
    # Put quotes around all values outside the no_quote_list
    def quotes_wrap(value, key):
        if key.lower() in lower_list(no_quote_list):
            return value
        return '"{}"'.format(value)

    jinja_map = dict_map(jinja_map, value_func=quotes_wrap)
    return jinja_map


def partition_templates(cmd, jinja_map):
    # Partition rendered attributes by parameters and attributes
    parameter_list = lower_list(cmdlet_parameters[cmd])
    other_attributes, parameters = dict_partition(
        lambda key, _: key.lower() in parameter_list, jinja_map
    )
    return parameters, other_attributes


def filter_illegal(cmd, parameters, other_attributes):

    # Drop all illegal parameters and attributes
    # Parameters
    for parameter in illegal_parameters[cmd]:
        parameters.pop(parameter, None)
        parameters.pop(parameter.lower(), None)
    # Attributes
    for attribute in illegal_attributes[cmd]:
        other_attributes.pop(attribute, None)
        other_attributes.pop(attribute.lower(), None)

    return parameters, other_attributes


def filter_empty_values(
    environment: Environment,
    attrs: Dict[str, str],
    context: Dict[str, Any],
) -> Dict[str, str]:
    """Remove key/template pairs from `attrs` if the template renders the value
    "\"None\"".
    """
    to_remove = set()

    for name, template_code in attrs.items():
        template = load_jinja_template(environment, template_code)
        value = template.render(**context)
        if value in ('"None"', "", f'"{INVALID}"'):
            to_remove.add(name)

    for attribute_name in to_remove:
        del attrs[attribute_name]

    return attrs


def load_jinja_template(environment: Environment, source: str) -> Template:
    """Load Jinja template in the string `source` and return a `Template`
    instance.
    """
    return environment.from_string(source)


def prepare_template(environment: Environment, cmd, settings, context):
    """Build a complete powershell command template.

    Args:
        environment: Jinja2 `Environment` instance
        cmd: command to generate template for.
        settings: dictionary containing settings from settings.json

    Returns:
        str: A jinja template string produced by templating the command
             template with all the field templates.
    """
    # Load command template via cmd
    cmd_options = cmdlet_templates.keys()
    if cmd not in cmd_options:
        raise ValueError(
            "prepare_template cmd must be one of: " + ",".join(cmd_options)
        )

    command_template = load_jinja_template(environment, cmdlet_templates[cmd])

    parameters, other_attributes = filter_illegal(
        cmd,
        *partition_templates(
            cmd, quote_templates(prepare_field_templates(cmd, settings))
        ),
    )

    parameters = filter_empty_values(environment, parameters, context)
    other_attributes = filter_empty_values(environment, other_attributes, context)

    # Generate our combined template, by rendering our command template using
    # the field templates templates.
    combined_template = command_template.render(
        parameters=parameters, other_attributes=other_attributes
    )
    return combined_template


def template_powershell(
    context,
    settings,
    cmd: str = "New-ADUser",
    environment: Environment = Environment(undefined=StrictUndefined),
) -> str:
    """Build a complete powershell command.

    Args:
        context: dictionary used for jinja templating context.
        settings: dictionary containing settings from settings.json
        cmd: command to generate template for. Defaults to 'New-ADUser'.
        environment: Jinja template environment

    Returns:
        str: An executable powershell script.
    """
    # Acquire the full template, templated itself with all field templates
    full_template = prepare_template(environment, cmd, settings, context)

    # Render the final template using the context
    final_template = load_jinja_template(environment, full_template)
    return final_template.render(**context)
